import json
import sys
from typing import Dict, List, Tuple, Callable, Union, Generator, Optional
from concurrent.futures import ThreadPoolExecutor, as_completed

import requests
import tqdm
import zulu
import json

import spyctl.cli as cli
import spyctl.spyctl_lib as lib
from spyctl.cache_dict import CacheDict

# Get policy parameters
GET_POL_TYPE = "type"
GET_POL_HAS_TAGS = "has_tags"
GET_POL_NAME_CONTAINS = "name_contains"
GET_POL_POLICY_CONTAINS = "policy_contains"
GET_POL_SELECTOR_CONTAINS = "selector_contains"
GET_POL_UID_EQUALS = "uid_equals"


# https://requests.readthedocs.io/en/latest/user/advanced/#timeouts
# connection timeout, read timeout
TIMEOUT = (30, 300)

AUTO_HIDE_TIME = zulu.now().shift(days=-1)
MAX_TIME_RANGE_SECS = 43200  # 12 hours
NAMESPACES_MAX_RANGE_SECS = 2000
TIMEOUT_MSG = "A timeout occurred during the API request. "


class NotFoundException(ValueError):
    pass


# ----------------------------------------------------------------- #
#                           API Primitives                          #
# ----------------------------------------------------------------- #


def get(url, key, params=None, raise_notfound=False):
    if key:
        headers = {
            "Authorization": f"Bearer {key}",
            "content-type": "application/json",
            "accept": "application/json",
        }
    else:
        headers = None
    try:
        r = requests.get(url, headers=headers, timeout=TIMEOUT, params=params)
    except requests.exceptions.Timeout as e:
        cli.err_exit(TIMEOUT_MSG + str(*e.args))
    context_uid = r.headers.get("x-context-uid", "No context uid found.")
    if lib.DEBUG:
        print(
            f"Request to {url}\n\tcontext_uid: {context_uid}"
            f"\n\tstatus: {r.status_code}"
        )
    if r.status_code == 404 and raise_notfound:
        raise NotFoundException()
    if r.status_code != 200:
        if "x-context-uid" in r.headers:
            context_uid = r.headers["x-context-uid"]
        else:
            context_uid = "No context uid found."
        msg = [f"{r.status_code}, {r.reason}", f"\tContext UID: {context_uid}"]
        if r.text:
            try:
                error = json.loads(r.text)
                if "msg" in error:
                    msg.append(error["msg"])
                else:
                    msg.append(f"{r.text}")
            except Exception:
                msg.append(f"{r.text}")
        msg = "\n".join(msg)
        cli.err_exit(msg)
    return r


def post(url, data, key, raise_notfound=False):
    headers = {"Authorization": f"Bearer {key}"}
    try:
        r = requests.post(url, json=data, headers=headers, timeout=TIMEOUT)
    except requests.exceptions.Timeout as e:
        cli.err_exit(TIMEOUT_MSG + str(e.args))
    context_uid = r.headers.get("x-context-uid", "No context uid found.")
    if lib.DEBUG:
        print(
            f"Request to {url}\n\tcontext_uid: {context_uid}"
            f"\n\tstatus: {r.status_code}"
        )
    if r.status_code == 404 and raise_notfound:
        raise NotFoundException()
    if r.status_code != 200:
        if "x-context-uid" in r.headers:
            context_uid = r.headers["x-context-uid"]
        else:
            context_uid = "No context uid found."
        msg = [f"{r.status_code}, {r.reason}", f"\tContext UID: {context_uid}"]
        if r.text:
            try:
                error = json.loads(r.text)
                if "msg" in error:
                    msg.append(error["msg"])
                else:
                    msg.append(f"{r.text}")
            except Exception:
                msg.append(f"{r.text}")
        msg = "\n".join(msg)
        cli.err_exit(msg)
    return r


def put(url, data, key):
    headers = {"Authorization": f"Bearer {key}"}
    try:
        r = requests.put(url, json=data, headers=headers, timeout=TIMEOUT)
    except requests.exceptions.Timeout as e:
        cli.err_exit(TIMEOUT_MSG + str(e.args))
    context_uid = r.headers.get("x-context-uid", "No context uid found.")
    if lib.DEBUG:
        print(
            f"Request to {url}\n\tcontext_uid: {context_uid}"
            f"\n\tstatus: {r.status_code}"
        )
    if r.status_code != 200:
        if "x-context-uid" in r.headers:
            context_uid = r.headers["x-context-uid"]
        else:
            context_uid = "No context uid found."
        msg = [f"{r.status_code}, {r.reason}", f"\tContext UID: {context_uid}"]
        if r.text:
            try:
                error = json.loads(r.text)
                if "msg" in error:
                    msg.append(error["msg"])
                else:
                    msg.append(f"{r.text}")
            except Exception:
                msg.append(f"{r.text}")
        msg = "\n".join(msg)
        cli.err_exit(msg)
    return r


def delete(url, key):
    headers = {"Authorization": f"Bearer {key}"}
    try:
        r = requests.delete(url, headers=headers, timeout=TIMEOUT)
    except requests.exceptions.Timeout as e:
        cli.err_exit(TIMEOUT_MSG + str(e.args))
    context_uid = r.headers.get("x-context-uid", "No context uid found.")
    if lib.DEBUG:
        print(
            f"Request to {url}\n\tcontext_uid: {context_uid}"
            f"\n\tstatus: {r.status_code}"
        )
    if r.status_code != 200:
        if "x-context-uid" in r.headers:
            context_uid = r.headers["x-context-uid"]
        else:
            context_uid = "No context uid found."
        msg = [f"{r.status_code}, {r.reason}", f"\tContext UID: {context_uid}"]
        if r.text:
            try:
                error = json.loads(r.text)
                if "msg" in error:
                    msg.append(error["msg"])
                else:
                    msg.append(f"{r.text}")
            except Exception:
                msg.append(f"{r.text}")
        msg = "\n".join(msg)
        cli.err_exit(msg)
    return r


# ----------------------------------------------------------------- #
#                     Main Retrieval Functions                      #
# ----------------------------------------------------------------- #


# Elastic Retrieval
def get_object_by_id(
    api_url,
    api_key,
    org_uid,
    id: str,
    schema: str,
    time: Tuple[float, float] = None,
    datatype="spydergraph",
):
    if time:
        hour_floor = lib.truncate_hour_epoch(time[0])
        data = {
            "data_type": datatype,
            "org_uid": org_uid,
            "query": f'(id:"{id}") AND (schema:"{schema}")'
            f" AND ((((valid_to:>{time[0]})"
            f' OR ((NOT status:"closed") AND time:>={hour_floor}))'
            f" AND valid_from:<{time[1]}) OR (time:[{time[0]} TO {time[1]}]"
            f" AND NOT _exists_:valid_to))",
            "query_from": 0,
            "query_size": 1,
        }
    else:
        data = {
            "data_type": datatype,
            "org_uid": org_uid,
            "query": f'(id:"{id}")',
            "query_from": 0,
            "query_size": 1,
        }
    url = (
        f"{api_url}/api/v1/source/query/"
        "?ui_tag=SearchLoadAllSchemaTypesInOneQuery"
    )
    return post(url, data, api_key)


DEFAULT_CACHE_DICT_LEN = 10000


# Source-based Retrieval
def retrieve_data(
    api_url: str,
    api_key: str,
    org_uid: str,
    sources: Union[str, List[str]],
    datatype: str,
    schema: str,
    time: Tuple[float, float],
    raise_notfound=False,
    pipeline=None,
    url="api/v1/source/query/",
    disable_pbar=False,
    limit_mem=True,
    disable_pbar_on_first=False,
):
    """This is the defacto data retrieval function. Most queries that don't
    target the SQL db can be executed with this function. It enforces limited
    memory usage unless told otherwise, shows a progress bar unless told
    otherwise, and yields records one at a time. The data returned is unsorted.

    Args:
        api_url (str): Top-most part of the API url -- from context
        api_key (str): Key to access the API -- from context
        org_uid (str): Org to get data from -- from context
        source (str, List[str]): The uid(s) that the data are tied to
        datatype (str): The data stream to look for the data in
        schema (str): A prefix of the schema for the desired objects
            (ex. model_connection, model_process)
        time (Tuple[float, float]): A tuple with (starting time, ending time)
        raise_notfound (bool, optional): Error to raise if the API throws an
            404 error. Defaults to False.
        pipeline (_type_, optional): Filtering done by the api.
            Defaults to None.
        url (_type_, optional): Alternative url path (ex. f"{api_url}/{url}").
            Defaults to "api/v1/source/query/".
        disable_pbar (bool, optional): Does not show the progress bar if set
            to True. Defaults to False.
        limit_mem (bool, optional): Limits the memory usage on the Latest
            Model calculation. If True we may return duplicate objects.
            Defaults to True.
        disable_pbar_on_first (bool, optional): Closes and clears the progress
            bar after first item is returned. Defaults to False.

    Yields:
        Iterator[dict]: An iterator over retrieved objects.
    """

    progress_bar_tracker: List[tqdm.tqdm] = []
    popped_off_cache = []

    def yield_on_del(_, value):
        if disable_pbar_on_first:
            progress_bar_tracker[0].close()
        popped_off_cache.append(value)

    cache_len = DEFAULT_CACHE_DICT_LEN if limit_mem else None
    data = CacheDict(cache_len=cache_len, on_del=yield_on_del)

    def new_version(id: str, obj: dict) -> bool:
        new_v = obj.get("version")
        if not new_v:
            return True
        old_obj: Dict = data.get(id)
        if not old_obj:
            return True
        old_v = old_obj.get("version")
        if not old_v:
            return True
        return new_v > old_v

    if isinstance(sources, str):
        sources = [sources]

    for resp in threadpool_progress_bar_time_blocks(
        sources,
        time,
        lambda src_uid, time_tup: get_filtered_data(
            api_url,
            api_key,
            org_uid,
            src_uid,
            datatype,
            schema,
            time_tup,
            raise_notfound,
            pipeline,
            url,
        ),
        disable_pbar=disable_pbar,
        pbar_tracker=progress_bar_tracker,
    ):
        if not resp:
            continue
        for json_obj in resp.iter_lines():
            obj = json.loads(json_obj)
            id = obj.get("id")
            if id:
                if new_version(id, obj):
                    data[id] = obj
                    while len(popped_off_cache) > 0:
                        yield popped_off_cache.pop()
            else:
                yield obj
    else:
        while len(data) > 0:
            yield data.popitem()[1]


def get_filtered_data(
    api_url,
    api_key,
    org_uid,
    source,
    datatype,
    schema,
    time,
    raise_notfound=False,
    pipeline=None,
    url="api/v1/source/query/",
) -> Optional[requests.Response]:
    """This function formats and makes a post request following the
    "source query" format. If a pipeline is not provided, this function will
    craft a basic one.

    Args:
        api_url (str): Top-most part of the API url -- from context
        api_key (str): Key to access the API -- from context
        org_uid (str): Org to get data from -- from context
        source (str, List[str]): The uid(s) that the data are tied to
        datatype (str): The data stream to look for the data in
        schema (str): A prefix of the schema for the desired objects
            (ex. model_connection, model_process)
        time (Tuple[float, float]): A tuple with (starting time, ending time)
        raise_notfound (bool, optional): Error to raise if the API throws an
            error. Defaults to False.
        pipeline (_type_, optional): Filtering done by the api.
            Defaults to None.
        url (_type_, optional): Alternative url path (ex. f"{api_url}/{url}").
            Defaults to "api/v1/source/query/".

    Returns:
        Response: The http response from the request.
    """
    url = f"{api_url}/{url}"
    data = {
        "start_time": time[0],
        "end_time": time[1],
        "data_type": datatype,
        "pipeline": [{"filter": {"schema": schema}}, {"latest_model": {}}],
    }
    if org_uid:
        data["org_uid"] = org_uid
    if pipeline:
        data["pipeline"] = pipeline
    if source:
        data["src_uid"] = source
    try:
        return post(url, data, api_key, raise_notfound)
    except NotFoundException:
        return None


def threadpool_progress_bar_time_blocks(
    args_per_thread: List[str],
    time,
    function: Callable,
    max_time_range=MAX_TIME_RANGE_SECS,
    disable_pbar=False,
    pbar_tracker: List = [],
) -> str:
    """This function runs a multi-threaded task such as making multiple API
    requests simultaneously. By default it shows a progress bar. This is a
    specialized function for the Spyderbat API because it will break up
    api-requests into time blocks of a maximum size if necessary. The
    Spyderbat API doesn't like queries spanning over 24 hours so we break them
    into smaller chunks.

    Args:
        args_per_thread (List[str]): The args to pass to each thread example:
            list of source uids.
        time (Tuple[float, float]): A tuple containing the start and end time
            of the task
        function (Callable): The function that each thread will perform
        max_time_range (_type_, optional): The maximum size of a time block.
            Defaults to MAX_TIME_RANGE_SECS.
        disable_pbar (bool, optional): Disable the progress bar.
            Defaults to False.
        pbar_tracker (list): A list that allows calling functions to control
            the pbar.

    Yields:
        Iterator[any]: The return value from the thread task.
    """
    t_blocks = time_blocks(time, max_time_range)
    args_per_thread = [
        [arg, t_block] for arg in args_per_thread for t_block in t_blocks
    ]
    pbar = tqdm.tqdm(
        total=len(args_per_thread),
        leave=False,
        file=sys.stderr,
        disable=disable_pbar,
    )
    pbar_tracker.clear()
    pbar_tracker.append(pbar)
    threads = []
    with ThreadPoolExecutor() as executor:
        for args in args_per_thread:
            threads.append(executor.submit(function, *args))
        for task in as_completed(threads):
            pbar.update(1)
            yield task.result()


def threadpool_progress_bar(
    args_per_thread: Union[List[List], List[str]],
    function: Callable,
    unpack_args=False,
):
    """A simplified version of the above function. In most cases it is
    best to use the time_blocks version unless you really know what you're
    doing.

    Args:
        args_per_thread (Union[List[List], List[str]]): The args to pass to
            each thread example:
            list of source uids.
        function (Callable): The function that each thread will perform
        unpack_args (bool, optional): _description_. Defaults to False.

    Yields:
        Iterator[any]: The return value from the thread task.
    """
    pbar = tqdm.tqdm(total=len(args_per_thread), leave=False, file=sys.stderr)
    threads = []
    with ThreadPoolExecutor() as executor:
        for args in args_per_thread:
            if unpack_args:
                threads.append(executor.submit(function, *args))
            else:
                threads.append(executor.submit(function, args))
        for task in as_completed(threads):
            pbar.update(1)
            yield task.result()


def time_blocks(
    time_tup: Tuple, max_time_range=MAX_TIME_RANGE_SECS
) -> List[Tuple]:
    """Takes a time tuple (start, end) in epoch time and converts
    it to smaller chunks if necessary.

    Args:
        time_tup (Tuple): start, end

    Returns:
        List[Tuple]: A list of (start, end) tuples to be used in api
            queries
    """
    st, et = time_tup
    if et - st > max_time_range:
        rv = []
        while et - st > max_time_range:
            et2 = min(et, st + max_time_range)
            rv.append((st, et2))
            st = et2
        rv.append((st, et))
        return rv
    else:
        return [time_tup]


# ----------------------------------------------------------------- #
#                        SQL-Based Resources                        #
# ----------------------------------------------------------------- #


def get_clusters(api_url, api_key, org_uid) -> List[Dict]:
    clusters = []
    url = f"{api_url}/api/v1/org/{org_uid}/cluster/"
    json = get(url, api_key).json()
    for cluster in json:
        if "/" not in cluster["uid"]:
            clusters.append(cluster)
    return clusters


def get_notification_policy(api_url, api_key, org_uid) -> Dict:
    url = f"{api_url}/api/v1/org/{org_uid}/notification_policy/"
    json = get(url, api_key).json()
    return json


def put_notification_policy(api_url, api_key, org_uid, notification_pol):
    url = f"{api_url}/api/v1/org/{org_uid}/notification_policy/"
    resp = put(url, notification_pol, api_key)
    return resp


def post_test_notification(api_url, api_key, org_uid, target_name):
    url = f"{api_url}/api/v1/org/{org_uid}/notification_policy/test_target"
    resp = post(url, {"target": target_name}, api_key)
    return resp


def get_sources(api_url, api_key, org_uid) -> List[Dict]:
    machines: Dict[str, Dict] = {}
    url = f"{api_url}/api/v1/org/{org_uid}/source/"
    source_json = get(url, api_key).json()
    for source in source_json:
        src_uid = source["uid"]
        if not src_uid.startswith("global"):
            machines[src_uid] = source
    # agents API call to find "description" (name used by the UI)
    url = f"{api_url}/api/v1/org/{org_uid}/agent/"
    agent_json = get(url, api_key).json()
    for agent in agent_json:
        src_uid = agent["runtime_details"]["src_uid"]
        description = agent["description"]
        if not agent["uid"].startswith("global"):
            source = machines.get(src_uid)
            if source is None:
                continue
            source.update(agent)
            machine = {}
            machine["uid"] = src_uid
            machine["name"] = description
            del source["uid"]
            del source["description"]
            del source["name"]
            machine.update(source)
            machines[src_uid] = machine
    # Auto-hide inactive machines
    rv = []
    for machine in machines.values():
        if (
            zulu.Zulu.parse(machine["last_data"]) >= AUTO_HIDE_TIME
            or zulu.Zulu.parse(machine["last_stored_chunk_end_time"])
            >= AUTO_HIDE_TIME
        ) and "runtime_details" in machine:
            rv.append(machine)
    return rv


def get_orgs(api_url, api_key) -> List[Tuple]:
    org_uids = []
    org_names = []
    url = f"{api_url}/api/v1/org/"
    orgs_json = get(url, api_key).json()
    for org in orgs_json:
        org_uids.append(org["uid"])
        org_names.append(org["name"])
    return (org_uids, org_names)


# ----------------------------------------------------------------- #
#                       Source-Based Resources                      #
# ----------------------------------------------------------------- #


def get_agents(
    api_url,
    api_key,
    org_uid,
    muids,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
) -> Generator[Dict, None, None]:
    try:
        datatype = lib.DATATYPE_AGENTS
        schema = lib.MODEL_AGENT_SCHEMA_PREFIX
        for agent in retrieve_data(
            api_url,
            api_key,
            org_uid,
            muids,
            datatype,
            schema,
            time,
            pipeline=pipeline,
            limit_mem=limit_mem,
            disable_pbar_on_first=disable_pbar_on_first,
        ):
            yield agent
    except KeyboardInterrupt:
        __log_interrupt()


def get_connections(
    api_url,
    api_key,
    org_uid,
    muids,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
) -> Generator[Dict, None, None]:
    try:
        datatype = lib.DATATYPE_SPYDERGRAPH
        schema = lib.MODEL_CONNECTION_PREFIX
        for connection in retrieve_data(
            api_url,
            api_key,
            org_uid,
            muids,
            datatype,
            schema,
            time,
            pipeline=pipeline,
            limit_mem=limit_mem,
            disable_pbar_on_first=disable_pbar_on_first,
        ):
            yield connection
    except KeyboardInterrupt:
        __log_interrupt()


def get_connection_bundles(
    api_url,
    api_key,
    org_uid,
    sources,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
):
    try:
        if sources and sources[0].startswith("clus:"):
            datatype = lib.DATATYPE_K8S
        else:
            datatype = lib.DATATYPE_SPYDERGRAPH
        schema = lib.MODEL_CONN_BUN_PREFIX
        for conn_bun in retrieve_data(
            api_url,
            api_key,
            org_uid,
            sources,
            datatype,
            schema,
            time,
            pipeline=pipeline,
            limit_mem=limit_mem,
            disable_pbar_on_first=disable_pbar_on_first,
        ):
            yield conn_bun
    except KeyboardInterrupt:
        __log_interrupt()


def get_containers(
    api_url,
    api_key,
    org_uid,
    sources,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
) -> Generator[Dict, None, None]:
    try:
        if sources and sources[0].startswith("clus"):
            datatype = lib.DATATYPE_K8S
        else:
            datatype = lib.DATATYPE_SPYDERGRAPH
        schema = lib.MODEL_CONTAINER_PREFIX
        for container in retrieve_data(
            api_url,
            api_key,
            org_uid,
            sources,
            datatype,
            schema,
            time,
            raise_notfound=True,
            pipeline=pipeline,
            limit_mem=limit_mem,
            disable_pbar_on_first=disable_pbar_on_first,
        ):
            yield container
    except KeyboardInterrupt:
        __log_interrupt()


def get_deployments(
    api_url,
    api_key,
    org_uid,
    clusters,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
) -> Generator[Dict, None, None]:
    try:
        datatype = lib.DATATYPE_K8S
        schema = lib.MODEL_DEPLOYMENT_PREFIX
        for deployment in retrieve_data(
            api_url,
            api_key,
            org_uid,
            clusters,
            datatype,
            schema,
            time,
            raise_notfound=True,
            pipeline=pipeline,
            limit_mem=limit_mem,
            disable_pbar_on_first=disable_pbar_on_first,
        ):
            yield deployment
    except KeyboardInterrupt:
        __log_interrupt()


def get_deviations(
    api_url,
    api_key,
    org_uid,
    policy_uids,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar=False,
    disable_pbar_on_first: bool = False,
) -> Generator[Dict, None, None]:
    try:
        datatype = lib.DATATYPE_AUDIT
        schema = (
            f"{lib.EVENT_AUDIT_PREFIX}:"
            f"{lib.EVENT_AUDIT_SUBTYPE_MAP['deviation']}"
        )
        url = f"api/v1/org/{org_uid}/analyticspolicy/logs"
        for deviation in retrieve_data(
            api_url,
            api_key,
            org_uid,
            policy_uids,
            datatype,
            schema,
            time,
            url=url,
            raise_notfound=True,
            pipeline=pipeline,
            limit_mem=limit_mem,
            disable_pbar=disable_pbar,
            disable_pbar_on_first=disable_pbar_on_first,
        ):
            yield deviation
    except KeyboardInterrupt:
        __log_interrupt()


def get_fingerprints(
    api_url,
    api_key,
    org_uid,
    sources,
    time,
    fprint_type=None,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
) -> Generator[Dict, None, None]:
    try:
        datatype = lib.DATATYPE_FINGERPRINTS
        if fprint_type:
            schema = (
                f"{lib.MODEL_FINGERPRINT_PREFIX}:"
                f"{lib.MODEL_FINGERPRINT_SUBTYPE_MAP[fprint_type]}"
            )
        else:
            schema = lib.MODEL_FINGERPRINT_PREFIX
        for fingerprint in retrieve_data(
            api_url,
            api_key,
            org_uid,
            sources,
            datatype,
            schema,
            time,
            pipeline=pipeline,
            limit_mem=limit_mem,
            disable_pbar_on_first=disable_pbar_on_first,
        ):
            if fingerprint.get("metadata", {}).get("type") not in {
                lib.POL_TYPE_CONT,
                lib.POL_TYPE_SVC,
            }:
                continue
            yield fingerprint
    except KeyboardInterrupt:
        __log_interrupt()


def get_machines(
    api_url,
    api_key,
    org_uid,
    muids,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
) -> Generator[Dict, None, None]:
    try:
        datatype = lib.DATATYPE_SPYDERGRAPH
        schema = lib.MODEL_MACHINE_PREFIX
        for machine in retrieve_data(
            api_url,
            api_key,
            org_uid,
            muids,
            datatype,
            schema,
            time,
            pipeline=pipeline,
            limit_mem=limit_mem,
            disable_pbar_on_first=disable_pbar_on_first,
        ):
            yield machine
    except KeyboardInterrupt:
        __log_interrupt()


def get_namespaces(
    api_url,
    api_key,
    org_uid,
    clusters,
    time,
    pipeline=None,
    disable_pbar_on_first: bool = False,
) -> Generator[Dict, None, None]:
    try:
        datatype = lib.DATATYPE_K8S
        schema = lib.MODEL_CLUSTER_PREFIX
        namespaces = list(
            retrieve_data(
                api_url,
                api_key,
                org_uid,
                clusters,
                datatype,
                schema,
                time,
                raise_notfound=True,
                pipeline=pipeline,
                limit_mem=False,
                disable_pbar_on_first=disable_pbar_on_first,
            )
        )
        for namespace in namespaces:
            yield namespace
    except KeyboardInterrupt:
        __log_interrupt()


def get_nodes(
    api_url,
    api_key,
    org_uid,
    clusters,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
) -> Generator[Dict, None, None]:
    try:
        datatype = lib.DATATYPE_K8S
        schema = lib.MODEL_NODE_PREFIX
        for node in retrieve_data(
            api_url,
            api_key,
            org_uid,
            clusters,
            datatype,
            schema,
            time,
            raise_notfound=True,
            pipeline=pipeline,
            limit_mem=limit_mem,
            disable_pbar_on_first=disable_pbar_on_first,
        ):
            yield node
    except KeyboardInterrupt:
        __log_interrupt()


def get_opsflags(
    api_url,
    api_key,
    org_uid,
    sources,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
) -> Generator[Dict, None, None]:
    try:
        if sources and sources[0].startswith("clus:"):
            datatype = lib.DATATYPE_K8S
        else:
            datatype = lib.DATATYPE_REDFLAGS
        schema = lib.EVENT_OPSFLAG_PREFIX
        for opsflag in retrieve_data(
            api_url,
            api_key,
            org_uid,
            sources,
            datatype,
            schema,
            time,
            pipeline=pipeline,
            limit_mem=limit_mem,
            disable_pbar_on_first=disable_pbar_on_first,
        ):
            yield opsflag
    except KeyboardInterrupt:
        __log_interrupt()


def get_pods(
    api_url,
    api_key,
    org_uid,
    clusters,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
) -> Generator[Dict, None, None]:
    try:
        datatype = lib.DATATYPE_K8S
        schema = lib.MODEL_POD_PREFIX
        for pod in retrieve_data(
            api_url,
            api_key,
            org_uid,
            clusters,
            datatype,
            schema,
            time,
            raise_notfound=True,
            pipeline=pipeline,
            limit_mem=limit_mem,
            disable_pbar_on_first=disable_pbar_on_first,
        ):
            yield pod
    except KeyboardInterrupt:
        __log_interrupt()


def get_processes(
    api_url,
    api_key,
    org_uid,
    muids,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
) -> Generator[Dict, None, None]:
    try:
        datatype = lib.DATATYPE_SPYDERGRAPH
        schema = lib.MODEL_PROCESS_PREFIX
        for process in retrieve_data(
            api_url,
            api_key,
            org_uid,
            muids,
            datatype,
            schema,
            time,
            pipeline=pipeline,
            limit_mem=limit_mem,
            disable_pbar_on_first=disable_pbar_on_first,
        ):
            yield process
    except KeyboardInterrupt:
        __log_interrupt()


def get_redflags(
    api_url,
    api_key,
    org_uid,
    sources,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
) -> Generator[Dict, None, None]:
    try:
        if sources and sources[0].startswith("clus:"):
            datatype = lib.DATATYPE_K8S
        else:
            datatype = lib.DATATYPE_REDFLAGS
        schema = lib.EVENT_REDFLAG_PREFIX
        for redflag in retrieve_data(
            api_url,
            api_key,
            org_uid,
            sources,
            datatype,
            schema,
            time,
            pipeline=pipeline,
            limit_mem=limit_mem,
            disable_pbar_on_first=disable_pbar_on_first,
        ):
            yield redflag
    except KeyboardInterrupt:
        __log_interrupt()


def get_spydertraces(
    api_url,
    api_key,
    org_uid,
    muids,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
) -> Generator[Dict, None, None]:
    try:
        datatype = lib.DATATYPE_SPYDERGRAPH
        schema = lib.MODEL_SPYDERTRACE_PREFIX
        for spydertrace in retrieve_data(
            api_url,
            api_key,
            org_uid,
            muids,
            datatype,
            schema,
            time,
            pipeline=pipeline,
            limit_mem=limit_mem,
            disable_pbar_on_first=disable_pbar_on_first,
        ):
            yield spydertrace
    except KeyboardInterrupt:
        __log_interrupt()


# ----------------------------------------------------------------- #
#                Policy Workflow SQL-Based Resources                #
# ----------------------------------------------------------------- #


def delete_policy(api_url, api_key, org_uid, pol_uid):
    url = f"{api_url}/api/v1/org/{org_uid}/analyticspolicy/{pol_uid}"
    resp = delete(url, api_key)
    return resp


def get_policies(api_url, api_key, org_uid, params=None):
    url = f"{api_url}/api/v1/org/{org_uid}/analyticspolicy/"
    params = {} if params is None else params
    if lib.METADATA_TYPE_FIELD in params:
        types = [params[lib.METADATA_TYPE_FIELD]]
    else:
        types = [lib.POL_TYPE_CONT, lib.POL_TYPE_SVC]
    policies = []
    for type in types:
        params[lib.METADATA_TYPE_FIELD] = type
        resp = get(url, api_key, params)
        for pol_json in resp.iter_lines():
            pol_list = json.loads(pol_json)
            for pol in pol_list:
                uid = pol["uid"]
                policy = json.loads(pol["policy"])
                policy[lib.METADATA_FIELD][lib.METADATA_UID_FIELD] = uid
                policy[lib.METADATA_FIELD][lib.METADATA_CREATE_TIME] = pol[
                    "valid_from"
                ]
                policies.append(policy)
    return policies


def get_policy(api_url, api_key, org_uid, pol_uid):
    url = f"{api_url}/api/v1/org/{org_uid}/analyticspolicy/{pol_uid}"
    resp = get(url, api_key)
    policies = []
    for pol_json in resp.iter_lines():
        pol = json.loads(pol_json)
        uid = pol["uid"]
        policy = pol["policy"]
        policy[lib.METADATA_FIELD][lib.METADATA_UID_FIELD] = uid
        policies.append(policy)
    return policies


def post_new_policy(api_url, api_key, org_uid, data: Dict):
    url = f"{api_url}/api/v1/org/{org_uid}/analyticspolicy/"
    resp = post(url, data, api_key)
    return resp


def put_policy_update(api_url, api_key, org_uid, pol_uid, data: Dict):
    url = f"{api_url}/api/v1/org/{org_uid}/analyticspolicy/{pol_uid}"
    resp = put(url, data, api_key)
    return resp


# ----------------------------------------------------------------- #
#               Policy Workflow Source-Based Resources              #
# ----------------------------------------------------------------- #


def get_audit_events(
    api_url,
    api_key,
    org_uid,
    time,
    src_uid,
    msg_type=None,
    since_id=None,
    disable_pbar: bool = False,
) -> List[Dict]:
    audit_events = []
    if msg_type:
        schema = (
            f"{lib.EVENT_AUDIT_PREFIX}:"
            f"{lib.EVENT_AUDIT_SUBTYPE_MAP[msg_type]}"
        )
    else:
        schema = lib.EVENT_AUDIT_PREFIX
    url = f"api/v1/org/{org_uid}/analyticspolicy/logs"
    for resp in threadpool_progress_bar_time_blocks(
        [src_uid],
        time,
        lambda uid, time_tup: get_filtered_data(
            api_url,
            api_key,
            org_uid,
            uid,
            "audit",
            schema,
            time_tup,
            pipeline=None,
            url=url,
        ),
        disable_pbar=disable_pbar,
    ):
        for event_json in resp.iter_lines():
            event = json.loads(event_json)
            audit_events.append(event)
    audit_events.sort(key=lambda event: event["time"])
    if since_id:
        for i, rec in enumerate(audit_events):
            if rec["id"] == since_id:
                if len(audit_events) > i + 1:
                    ind = i + 1
                    audit_events = audit_events[ind:]
                    break
                else:
                    return []
    return audit_events


def get_trace_summaries(
    api_url,
    api_key,
    org_uid,
    sources,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
) -> Generator[Dict, None, None]:
    try:
        datatype = lib.DATATYPE_FINGERPRINTS
        schema = f"{lib.MODEL_FINGERPRINT_PREFIX}:{lib.POL_TYPE_TRACE}"
        for fingerprint in retrieve_data(
            api_url,
            api_key,
            org_uid,
            sources,
            datatype,
            schema,
            time,
            pipeline=pipeline,
            limit_mem=limit_mem,
            disable_pbar_on_first=disable_pbar_on_first,
        ):
            yield fingerprint
    except KeyboardInterrupt:
        __log_interrupt()


# ----------------------------------------------------------------- #
#                          Aggregation Counts                       #
# ----------------------------------------------------------------- #


def deviation_count():
    pass


# ----------------------------------------------------------------- #
#                         Alternative Outputs                       #
# ----------------------------------------------------------------- #


def get_agent_metrics(
    api_url,
    api_key,
    org_uid,
    sources,
    time,
    pipeline=None,
    limit_mem: bool = False,
    disable_pbar_on_first: bool = False,
) -> Generator[Dict, None, None]:
    try:
        datatype = lib.DATATYPE_AGENTS
        schema = lib.EVENT_METRICS_PREFIX
        for metric in retrieve_data(
            api_url,
            api_key,
            org_uid,
            sources,
            datatype,
            schema,
            time,
            pipeline=pipeline,
            limit_mem=limit_mem,
            disable_pbar_on_first=disable_pbar_on_first,
        ):
            yield metric
    except KeyboardInterrupt:
        __log_interrupt()


def get_audit_events_tail(
    api_url,
    api_key,
    org_uid,
    time,
    src_uid,
    tail: int = -1,  # -1 means all events
    msg_type=None,
    since_id: str = None,
    disable_pbar: bool = False,
) -> List[Dict]:
    audit_events = get_audit_events(
        api_url,
        api_key,
        org_uid,
        time,
        src_uid,
        msg_type,
        since_id,
        disable_pbar=disable_pbar,
    )
    if tail > 0:
        return audit_events[-tail:]
    if tail == 0:
        return []
    else:
        return audit_events


def get_latest_agent_metrics(
    api_url,
    api_key,
    org_uid,
    args: List[Tuple[str, Tuple]],  # list (source_uid, (st, et))
    pipeline=None,
):
    try:
        for resp in threadpool_progress_bar(
            args,
            lambda source, time_tup: get_filtered_data(
                api_url,
                api_key,
                org_uid,
                source,
                lib.DATATYPE_AGENTS,
                lib.EVENT_METRICS_PREFIX,
                time_tup,
                pipeline=pipeline,
            ),
            unpack_args=True,
        ):
            latest_time = 0
            for json_obj in reversed(list(resp.iter_lines())):
                metrics_record = json.loads(json_obj)
                time = metrics_record["time"]
                if time <= latest_time:
                    break
                else:
                    latest_time = time
                yield metrics_record
    except KeyboardInterrupt:
        __log_interrupt()


def get_pypi_version():
    url = "https://pypi.org/pypi/spyctl/json"
    try:
        resp = get(url, key=None, raise_notfound=True)
        version = resp.json().get("info", {}).get("version")
        if not version:
            # cli.try_log("Unable to parse latest pypi version")
            return None
        return version
    except ValueError:
        # cli.try_log("Unable to reach version API")
        pass


def get_sources_data_for_agents(api_url, api_key, org_uid) -> Dict:
    rv = {}
    url = f"{api_url}/api/v1/org/{org_uid}/source/"
    sources = get(url, api_key).json()
    for source in sources:
        source_uid = source["uid"]  # muid
        if "runtime_details" not in source:
            rv[source_uid] = {
                "uid": source["uid"],
                "cloud_region": lib.NOT_AVAILABLE,
                "cloud_type": lib.NOT_AVAILABLE,
                "last_data": source["last_data"],
            }
        else:
            rv[source_uid] = {
                "uid": source["uid"],
                "cloud_region": source["runtime_details"].get(
                    "cloud_region", lib.NOT_AVAILABLE
                ),
                "cloud_type": source["runtime_details"].get(
                    "cloud_type", lib.NOT_AVAILABLE
                ),
                "last_data": source["last_data"],
            }
    return rv


def validate_search_query(
    api_url, api_key, org_uid, schema_type: str, query: str
):
    url = f"{api_url}/api/v1/org/{org_uid}/search/validate"
    data = {
        "context_uid": lib.build_ctx(),
        "query": query,
        "schema": schema_type,
    }
    resp = post(url, data, api_key)
    resp = resp.json()
    if not resp["ok"]:
        return resp["error"]
    else:
        return ""


# ----------------------------------------------------------------- #
#                          Helper Functions                         #
# ----------------------------------------------------------------- #


def __log_interrupt():
    cli.try_log("\nRequest aborted, no partial results.. exiting.")
    exit(0)


# ----------------------------------------------------------------- #
#                          Tester Functions                         #
# ----------------------------------------------------------------- #


def api_diff(api_url, api_key, org_uid, obj, d_objs):
    url = f"{api_url}/api/v1/org/{org_uid}/spyctl/diff/"
    data = {"diff_objects": json.dumps([d_objs]), "object": json.dumps(obj)}
    resp = post(url, data, api_key)
    diff_data = resp.json()["diff_data"]
    return diff_data


def api_create_guardian_policy(
    api_url, api_key, org_uid, name, mode, data: Dict
) -> str:
    url = f"{api_url}/api/v1/org/{org_uid}/spyctl/guardianpolicy/build/"
    data = {"input_objects": json.dumps([data]), "mode": mode}
    if name:
        data["name"] = name
    resp = post(url, data, api_key)
    policy = resp.json()["policy"]
    return policy


def api_create_suppression_policy(
    api_url,
    api_key,
    org_uid,
    name,
    type,
    scope_to_users,
    object_uid,
    **selectors,
) -> str:
    url = f"{api_url}/api/v1/org/{org_uid}/spyctl/suppressionpolicy/build/"
    data = {"type": type}

    def dash(key: str) -> str:
        return key.replace("_", "-")

    processed_selectors = {dash(k): v for k, v in selectors.items()}
    if name:
        data["name"] = name
    if scope_to_users:
        data["scope_to_users"] = scope_to_users
    if object_uid:
        data["object_uid"] = object_uid
    if processed_selectors:
        data["selectors"] = processed_selectors
    print(data)
    resp = post(url, data, api_key)
    policy = resp.json()["policy"]
    return policy


def api_merge(api_url, api_key, org_uid, obj, m_objs):
    url = f"{api_url}/api/v1/org/{org_uid}/spyctl/merge/"
    data = {"merge_objects": json.dumps([m_objs]), "object": json.dumps(obj)}
    resp = post(url, data, api_key)
    merged_object = resp.json()["merged_object"]
    return merged_object


def api_validate(api_url, api_key, org_uid, data: Dict) -> str:
    url = f"{api_url}/api/v1/org/{org_uid}/spyctl/validate/"
    resp = post(url, data, api_key)
    invalid_msg = resp.json()["invalid_message"]
    return invalid_msg



#!/usr/bin/env python3


"""
    This program prints a CSV file with the statistics of CPU, memory, and
    bandwidth utilization across a fleet of agents. 

    Fetch the data using, for example,
         spyctl get agents --usage-json -t 24h > usage.json

    Generate the CSV using
         ./agent.py -i usage.json > usage.csv

"""

import json
import argparse
import sys
import gzip
import math
from collections import defaultdict
import traceback


# -------------------------------------------------------------
def _parse_cmd_line():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument("-i", "--infile", required=True)
    args = parser.parse_args()
    return args


# -------------------------------------------------------------

options = _parse_cmd_line()

# Read the infile and print if it's one we're looking for
if options.infile.endswith("gz"):
    infile = gzip.open(options.infile, "rt")
else:
    infile = open(options.infile, "r")


def percentile(data, q):
    try:
        data_sorted = sorted(data)
        index = math.ceil(q / 100 * len(data_sorted))
        if index >= len(data):
            index = len(data) - 1
        return data_sorted[index]
    except IndexError:
        breakpoint()


agent = {}
mem = defaultdict(list)
cpu = defaultdict(list)
bandwidth = defaultdict(list)
try:
    for line in infile:
        rec = json.loads(line)
        agent_id = rec["agent_id"]
        agent[agent_id] = rec
        mem[agent_id].append(rec["mem_usage_B"])
        cpu[agent_id].append(rec["cpu_usage_1min"])
        bandwidth[agent_id].append(rec["bandwidth_1min_Bps"])
except Exception as exc:
    traceback.print_exc()
    print(exc)
infile.close()

print(f"agent,cpu mean,cpu p90,cpu p99,cpu max,mem mean,mem p90,mem p99,mem max,bps mean,bps p90,bps p99,bps max")
for agent_id in cpu:
    name = agent[agent_id]["agent_name"]
    w = [name]
    for label, table in [
        ("CPU", cpu),
        ("Mem", mem),
        ("Bps", bandwidth),
    ]:
        data = table[agent_id]
        mean = sum(data) / len(data)
        p90 = percentile(data, 90)
        p99 = percentile(data, 99)
        largest = max(data)
        w.extend([repr(mean), repr(p90), repr(p99), repr(largest)])
    print(','.join(w))



import os
import shutil
import unittest.mock as mock
from fnmatch import fnmatch
from itertools import groupby
from pathlib import Path

import pytest
from click.testing import CliRunner

import spyctl.tests.mock_functions as mock_func
from spyctl import spyctl
from spyctl.config.configs import CURR_CONTEXT_NONE, set_testing
from spyctl.spyctl_lib import time_inp

API_KEY = os.environ.get("API_KEY")
API_URL = os.environ.get("API_URL")
ORG = os.environ.get("ORG")
TEST_SECRET = "__test_secret__"
TEST_CONTEXT = "__test_context__"
CURR_PATH = Path.cwd()
WORKSPACE_DIR_NAME = "test_workspace__"
WORKSPACE_PATH = Path(f"./{WORKSPACE_DIR_NAME}")
WORKSPACE_CONFIG = Path(str(WORKSPACE_PATH.absolute()) + "/.spyctl/config")
CURRENT_CONTEXT = None


class SetupException(Exception):
    pass


def test_version():
    runner = CliRunner()
    response = runner.invoke(spyctl.main, ["-v"])
    test_pat = "Spyctl, version *.*.*"
    assert response.exit_code == 0
    assert fnmatch(response.output, test_pat)
    response = runner.invoke(spyctl.main, ["--version"])
    test_pat = "Spyctl, version *.*.*"
    assert response.exit_code == 0
    assert fnmatch(response.output, test_pat)


def test_help():
    runner = CliRunner()
    response = runner.invoke(spyctl.main, ["-h"])
    assert response.exit_code == 0
    response = runner.invoke(spyctl.main, ["--help"])
    assert response.exit_code == 0
    response = runner.invoke(spyctl.main, ["get", "redflags", "--help"])
    assert response.exit_code == 0
    assert "Options for redflags:" in response.output
    response = runner.invoke(spyctl.main, ["config", "--help"])
    assert response.exit_code == 0


@mock.patch(
    "spyctl.commands.get.api.get_namespaces", mock_func.mock_get_namespaces
)
def test_get_namespaces():
    get_resource("namespaces", TWOHOURS)
    get_resource("namespaces", TWOHOURS + OUTYML)


TWOHOURS = ("-t", "2h")
OUTYML = ("-o", "yaml")
MINS_30 = ("-t", "30m")
MINS_5 = ("-t", "5m")


def remove_timestamps(str_list):
    rv = []
    for string in str_list:
        try:
            time_inp(string)
        except ValueError:
            rv.append(string)
    return rv


# asserts that the table has some population and returns
# name_or_id values that should give the first line again
def process_table_response(output: str):
    name_or_id_fields = ("NAME", "UID", "FLAG", "IMAGE", "CGROUP")
    lines = output.strip().splitlines()
    assert len(lines) > 1
    header_line, first_line = lines[:2]
    headers = []
    for non_empty_group, chars in groupby(
        enumerate(header_line), lambda x: not x[1].isspace()
    ):
        if non_empty_group:
            pos, header = next(chars)
            header += "".join(c for _, c in chars)
            headers.append((pos, header))
    rv = []
    for i, (pos, header) in enumerate(headers):
        if header in name_or_id_fields:
            next_pos = headers[i + 1][0] if i < len(headers) - 1 else None
            table_element = first_line[pos:next_pos].strip()
            rv.append(table_element)
    print(header_line)
    print(headers)
    assert len(rv) > 0
    comparison_line = remove_timestamps(first_line.split())
    return comparison_line, rv


resources = (
    "clusters",
    "sources",
    "policies",
    "pods",
    "nodes",
    "deployments",
    "fingerprints",
    "namespaces",  # doesn't output a table
    "redflags",
    "opsflags",
)


@mock.patch.multiple(
    "spyctl.commands.get.api",
    get_clusters=mock_func.mock_get_clusters,
    get_deployments=mock_func.mock_get_deployments,
    get_fingerprints=mock_func.mock_get_fingerprints,
    get_namespaces=mock_func.mock_get_namespaces,
    get_nodes=mock_func.mock_get_nodes,
    get_opsflags=mock_func.mock_get_opsflags,
    get_pods=mock_func.mock_get_pods,
    get_policies=mock_func.mock_get_policies,
    get_redflags=mock_func.mock_get_redflags,
    get_sources=mock_func.mock_get_sources,
)
@mock.patch.multiple(
    "spyctl.resources.deployments.api",
    get_deployments=mock_func.mock_get_deployments,
)
@mock.patch.multiple(
    "spyctl.resources.flags.api",
    get_opsflags=mock_func.mock_get_opsflags,
    get_redflags=mock_func.mock_get_redflags,
)
@mock.patch.multiple(
    "spyctl.resources.namespaces.api",
    get_namespaces=mock_func.mock_get_namespaces,
)
@mock.patch.multiple(
    "spyctl.resources.nodes.api",
    get_nodes=mock_func.mock_get_nodes,
)
@mock.patch.multiple(
    "spyctl.resources.pods.api",
    get_pods=mock_func.mock_get_pods,
)
@pytest.mark.parametrize("resource", resources)
def test_get_resources(resource):
    if resource not in ["opsflags", "fingerprints"]:
        time_range = MINS_30
    else:
        time_range = TWOHOURS
    output = OUTYML
    table = get_resource(resource, time_range)
    comparison_line, ids = process_table_response(table)
    for name_or_id in ids:
        single = get_resource(resource, time_range + (name_or_id,))
        first_output = remove_timestamps(single.splitlines()[1].split())
        assert first_output == comparison_line
    time_range = MINS_5
    get_resource(resource, time_range + output)


def get_resource(resource, args=[], print_output=False):
    runner = CliRunner(mix_stderr=False)
    response = runner.invoke(spyctl.main, ["get", resource, *args])
    if print_output:
        print(response.output)
    assert response.exit_code == 0
    return response.stdout


resources_dir = Path(__file__).parent / "test_resources"


def test_create():
    runner = CliRunner()
    response = runner.invoke(
        spyctl.main,
        ["create", "policy", "-f", resources_dir / "test_baseline.yaml"],
    )
    assert response.exit_code == 0
    with open(resources_dir / "test_policy.yaml", "r") as f:
        assert response.output.strip("\n") == f.read().strip("\n")
    response = runner.invoke(
        spyctl.main,
        ["create", "baseline", "-f", resources_dir / "test_fprint_group.yaml"],
    )
    assert response.exit_code == 0
    with open(resources_dir / "test_baseline.yaml", "r") as f:
        assert response.output.strip("\n") == f.read().strip("\n")


@mock.patch(
    "spyctl.commands.get.api.get_policies", mock_func.mock_get_policies
)
@mock.patch.multiple(
    "spyctl.commands.apply.api",
    get_policies=mock_func.mock_get_policies,
    delete_policy=mock_func.mock_delete_policy,
    post_new_policy=mock_func.mock_post_new_policy,
    put_policy_update=mock_func.mock_put_policy_update,
)
@mock.patch.multiple(
    "spyctl.commands.delete.api",
    get_policies=mock_func.mock_get_policies,
    delete_policy=mock_func.mock_delete_policy,
)
def test_update_policy():
    runner = CliRunner()
    response = runner.invoke(
        spyctl.main, ["apply", "-f", resources_dir / "test_policy.yaml"]
    )
    assert response.exit_code == 0
    assert response.output.startswith(
        "Successfully applied new policy with uid:"
    )
    response = runner.invoke(
        spyctl.main, ["get", "policies", "spyderbat-test"]
    )
    assert response.exit_code == 0
    assert "spyderbat-test" in response.output
    lines = response.stdout.splitlines()
    uid = lines[2].split(" ")[0]
    get_runner = CliRunner(mix_stderr=False)
    response = get_runner.invoke(
        spyctl.main, ["get", "policies", uid, "-o", "yaml"]
    )
    assert response.exit_code == 0
    assert "spyderbat-test" in response.output
    policy_with_uid = "uid_pol.yaml"
    with open(policy_with_uid, "w") as f:
        f.write(response.output)
    response = runner.invoke(spyctl.main, ["apply", "-f", policy_with_uid])
    assert response.exit_code == 0
    assert response.output.startswith("Successfully updated policy")
    response = runner.invoke(
        spyctl.main, ["delete", "policy", "-y", "spyderbat-test"]
    )
    assert response.exit_code == 0
    assert response.output.startswith("Successfully deleted policy")


@mock.patch(
    "spyctl.commands.get.api.get_policies", mock_func.mock_get_policies
)
@mock.patch.multiple(
    "spyctl.commands.apply.api",
    get_policies=mock_func.mock_get_policies,
    delete_policy=mock_func.mock_delete_policy,
    post_new_policy=mock_func.mock_post_new_policy,
)
@mock.patch.multiple(
    "spyctl.commands.delete.api",
    get_policies=mock_func.mock_get_policies,
    delete_policy=mock_func.mock_delete_policy,
)
def test_apply_delete():
    runner = CliRunner()
    response = runner.invoke(
        spyctl.main, ["apply", "-f", Path(resources_dir, "test_policy.yaml")]
    )
    assert response.exit_code == 0
    assert response.output.startswith(
        "Successfully applied new policy with uid:"
    )
    response = runner.invoke(
        spyctl.main, ["get", "policies", "spyderbat-test"]
    )
    assert response.exit_code == 0
    assert "spyderbat-test" in response.output
    response = runner.invoke(
        spyctl.main, ["delete", "policy", "-y", "spyderbat-test"]
    )
    assert response.exit_code == 0
    assert response.output.startswith("Successfully deleted policy")


def test_diff():
    runner = CliRunner()
    response = runner.invoke(
        spyctl.main,
        [
            "diff",
            "-f",
            str(resources_dir) + "/test_baseline.yaml",
            "-w",
            str(resources_dir) + "/test_baseline_extra.yaml",
            "-y",
        ],
    )
    assert response.exit_code == 0
    assert "+     - name: python3.7" in response.output


def test_merge():
    runner = CliRunner()
    response = runner.invoke(
        spyctl.main,
        [
            "merge",
            "-f",
            str(resources_dir) + "/test_baseline.yaml",
            "-w",
            str(resources_dir) + "/test_baseline_extra.yaml",
            "-y",
        ],
    )
    assert response.exit_code == 0
    assert "\n    - name: python3.7" in response.output
    assert "\n    - name: sh" in response.output


def env_setup() -> bool:
    # Tests that the proper environment variables exist in pyproject.toml
    if not API_KEY or API_KEY == "__NONE__":
        print(
            "No api key provided. Edit the API_KEY environment variable in"
            " pyproject.toml"
        )
        return False
    if not API_URL or API_URL == "__NONE__":
        print(
            "No api url provided. Edit the API_URL environment variable in"
            ' pyproject.toml to "API_URL=https://api.spyderbat.com"'
        )
        return False
    if not ORG or ORG == "__NONE__":
        print(
            "No organization provided. Edit the ORG environment variable in"
            " pyproject.toml"
        )
        return False
    return True


def create_secret():
    runner = CliRunner()
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "set-apisecret",
            "-k",
            API_KEY,
            "-u",
            API_URL,
            TEST_SECRET,
        ],
    )
    if result.exit_code != 0:
        raise SetupException("Unable to create test secret")


def delete_secret():
    runner = CliRunner()
    runner.invoke(
        spyctl.main,
        [
            "config",
            "delete-apisecret",
            "-y",
            TEST_SECRET,
        ],
    )


def create_context(name):
    runner = CliRunner()
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "set-context",
            "-s",
            TEST_SECRET,
            "-o",
            ORG,
            "-g",
            name,
        ],
    )
    if result.exit_code != 0:
        raise SetupException("Unable to create test context")


def delete_context(name):
    runner = CliRunner()
    runner.invoke(
        spyctl.main,
        ["config", "delete-context", name],
    )


def delete_test_policies():
    runner = CliRunner()
    runner.invoke(spyctl.main, ["delete", "policy", "-y", "spyderbat-test"])


def current_context():
    runner = CliRunner()
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "current-context",
        ],
    )
    if result.exit_code != 0:
        raise SetupException("Unable get current context")
    current_ctx = result.output.strip("\n")
    return current_ctx


def use_test_context(name):
    runner = CliRunner()
    runner.invoke(spyctl.main, ["config", "use-context", name])


def use_current_context():
    if CURRENT_CONTEXT:
        runner = CliRunner()
        runner.invoke(spyctl.main, ["config", "use-context", CURRENT_CONTEXT])


def setup_module():
    if not env_setup():
        raise SetupException("Check environment variables in pyproject.toml")
    global CURRENT_CONTEXT
    current_ctx = current_context()
    if current_ctx and current_ctx != CURR_CONTEXT_NONE:
        CURRENT_CONTEXT = current_ctx
    if WORKSPACE_PATH.exists():
        teardown_module()
        raise SetupException("Workspace path already exists")
    WORKSPACE_PATH.mkdir(exist_ok=False)
    os.chdir(WORKSPACE_PATH)
    create_secret()
    create_context(TEST_CONTEXT)
    use_test_context(TEST_CONTEXT)
    # Set a global variable in configs.py to force reloading
    # the config file. Otherwise the global variable LOADED_CONFIG
    # remains set across invocations
    set_testing()


def teardown_module():
    delete_test_policies()
    delete_context(TEST_CONTEXT)
    delete_secret()
    use_current_context()
    os.chdir(CURR_PATH)
    shutil.rmtree(WORKSPACE_PATH, ignore_errors=True)



import spyctl.schemas_v2 as schemas
from fastapi import APIRouter, BackgroundTasks
from pydantic import BaseModel, Field, Json

import app.app_lib as app_lib
import app.commands.validate as cmd_val

router = APIRouter(prefix="/api/v1")

# ------------------------------------------------------------------------------
# Validate Spyderbat Object
# ------------------------------------------------------------------------------


class ValidateHandlerInput(BaseModel):
    object: Json[schemas.SpyderbatObject] = Field(
        title="The object to be validated."
    )


class ValidateHandlerOutput(BaseModel):
    invalid_message: str = Field(
        title="Message if the object is invalid,"
        " empty string if object is valid"
    )


@router.post("/validate")
def validate(
    i: ValidateHandlerInput, background_tasks: BackgroundTasks
) -> ValidateHandlerOutput:
    background_tasks.add_task(app_lib.flush_spyctl_log_messages)
    validate_input = cmd_val.ValidateInput(i.object.dict(by_alias=True))
    output = cmd_val.validate(validate_input)
    return ValidateHandlerOutput(invalid_message=output.invalid_message)



from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import click
import yaml
from click.shell_completion import CompletionItem
from tabulate import tabulate

import spyctl.api as api
import spyctl.cli as cli
import spyctl.config.secrets as s
import spyctl.spyctl_lib as lib
import spyctl.schemas_v2 as schemas

APP_NAME = "spyctl"
APP_DIR = f".{APP_NAME}"
APP_AUTHOR = "Spyderbat"
SECRETS_DIR = ".secrets"
GLOBAL_CONFIG_DIR = Path.joinpath(Path.home(), APP_DIR)
GLOBAL_SECRETS_DIR = Path.joinpath(GLOBAL_CONFIG_DIR, SECRETS_DIR)
CONFIG_FILENAME = "config"
SECRETS_FILENAME = "secrets"
GLOBAL_CONFIG_PATH = Path.joinpath(GLOBAL_CONFIG_DIR, CONFIG_FILENAME)
GLOBAL_SECRETS_PATH = Path.joinpath(GLOBAL_SECRETS_DIR, SECRETS_FILENAME)
LOCAL_CONFIG_PATH = Path(f"./{APP_DIR}/{CONFIG_FILENAME}")
LOCAL_SECRETS_PATH = Path(f"./{APP_DIR}/{SECRETS_DIR}/{SECRETS_FILENAME}")
CURRENT_CONTEXT = None
LOADED_CONFIG: "Config" = None
# Pa
LOADED_CONFIGS: Dict[str, "Config"] = {}

TESTING = False

# Config yaml required values
CONFIG_KIND = lib.CONFIG_KIND

# Config yaml fields
CURR_CONTEXT_FIELD = lib.CURR_CONTEXT_FIELD
CONTEXTS_FIELD = lib.CONTEXTS_FIELD
CONTEXT_FIELD = lib.CONTEXT_FIELD
CONTEXT_NAME_FIELD = lib.CONTEXT_NAME_FIELD
SECRET_FIELD = lib.SECRET_FIELD
ORG_FIELD = lib.ORG_FIELD
POD_FIELD = lib.POD_FIELD
CLUSTER_FIELD = lib.CLUSTER_FIELD
NAMESPACE_FIELD = lib.NAMESPACE_FIELD
MACHINES_FIELD = lib.MACHINES_FIELD
CGROUP_FIELD = lib.CGROUP_FIELD
CONTAINER_NAME_FIELD = lib.CONT_NAME_FIELD
CONT_ID_FIELD = lib.CONT_ID_FIELD
IMG_FIELD = lib.IMAGE_FIELD
IMGID_FIELD = lib.IMAGEID_FIELD

CURR_CONTEXT_NONE = "__none__"
CONFIG_TEMPLATE = {
    lib.API_FIELD: lib.API_VERSION,
    lib.KIND_FIELD: CONFIG_KIND,
    CONTEXTS_FIELD: [],
    CURR_CONTEXT_FIELD: CURR_CONTEXT_NONE,
}


class InvalidContextDataError(Exception):
    pass


class InvalidConfigDataError(Exception):
    pass


class Context:
    required_keys = {CONTEXT_NAME_FIELD, CONTEXT_FIELD, SECRET_FIELD}
    required_context_keys = {ORG_FIELD}
    optional_context_keys = {
        CLUSTER_FIELD,
        NAMESPACE_FIELD,
        MACHINES_FIELD,
        POD_FIELD,
        CGROUP_FIELD,
        CONTAINER_NAME_FIELD,
        CONT_ID_FIELD,
        IMG_FIELD,
        IMGID_FIELD,
    }
    required_creds = {lib.API_KEY_FIELD, lib.API_URL_FIELD}

    def __init__(self, context_data: Dict) -> None:
        if not isinstance(context_data, dict):
            raise InvalidContextDataError("Context should be a dictionary.")
        for key in self.required_keys:
            if key not in context_data:
                raise InvalidContextDataError(f"Context missing {key} field.")
        self.context = context_data[CONTEXT_FIELD]
        if not isinstance(self.context, dict):
            raise InvalidContextDataError(
                "Context field of a context should be a dictionary."
            )
        for key in self.required_context_keys:
            if key not in self.context:
                raise InvalidContextDataError(f"Context missing {key} field.")
        for key in self.required_context_keys.union(
            self.optional_context_keys
        ):
            if key in self.context and not (
                isinstance(self.context[key], str)
                or isinstance(self.context[key], list)
            ):
                raise InvalidContextDataError(
                    f"Value for {key} must be a string or a list"
                )
        self.name = context_data[CONTEXT_NAME_FIELD]
        if not isinstance(self.name, str) or not self.name:
            raise InvalidContextDataError(
                "Context name is not a valid string."
            )
        self.secret_name = context_data[SECRET_FIELD]
        if not isinstance(self.secret_name, str):
            raise InvalidContextDataError(
                "Context secret is not a valid string."
            )
        self.secret = s.find_secret(self.secret_name)
        self.org_uid = None
        self.filters = {}
        self.__set_filters()

    @property
    def global_source(self):
        self.get_api_data()
        return f"global:{self.org_uid}"

    def get_api_data(self) -> Tuple[str, str, str]:
        if self.secret is None:
            cli.err_exit(f"Unable to locate secret '{self.secret_name}'")
        secret_creds = self.secret.get_credentials()
        api_url = secret_creds[lib.API_URL_FIELD]
        api_key = secret_creds[lib.API_KEY_FIELD]
        if self.org_uid is None:
            self.org_uid = validate_org(self.org, api_url, api_key)
        if self.org_uid is None:
            cli.err_exit(
                "Unable to validate organization name or ID."
                " Check context settings. The api key in the"
                f" {lib.SECRET_KIND} {self.secret_name!r} must be"
                f" allowed to access the organization {self.org!r},"
                " and the organization name or uid must be correct."
                " As a last resort, check to see if the API key has expired."
            )
        return (api_url, api_key, self.org_uid)

    def get_filters(self) -> Dict[str, List[str]]:
        return self.filters

    def as_dict(self) -> Dict:
        rv = {
            CONTEXT_NAME_FIELD: self.name,
            SECRET_FIELD: self.secret_name,
            CONTEXT_FIELD: self.context,
        }
        return rv

    def __set_filters(self):
        if len(self.filters) == 0:
            self.filters = self.context

    def __repr__(self) -> str:
        return f'Context("{self.name}", "{self.org}")'

    @property
    def org(self) -> str:
        return self.context[ORG_FIELD]


class Config:
    required_keys = {
        lib.API_FIELD,
        lib.KIND_FIELD,
        CONTEXTS_FIELD,
        CURR_CONTEXT_FIELD,
    }
    validation_errors = []

    def __init__(
        self, config_data: Dict, config_path: Path, config_dir: Path
    ) -> None:
        self.config_path = config_path
        # config_dir identifies the location of the current workspace
        # is updated when more-local config files are merged in
        self.config_dir = config_dir
        for key in self.required_keys:
            if key not in config_data:
                raise InvalidConfigDataError(f"Config missing {key} field")
        if not lib.valid_api_version(config_data.get(lib.API_FIELD)):
            raise InvalidConfigDataError("Invalid apiVersion")
        if not lib.valid_kind(config_data.get(lib.KIND_FIELD), CONFIG_KIND):
            raise InvalidConfigDataError("Invalid kind")
        if not isinstance(config_data[CURR_CONTEXT_FIELD], str):
            raise InvalidConfigDataError(
                "Invalid current context, should be a string"
            )
        if not isinstance(config_data[CONTEXTS_FIELD], list):
            raise InvalidConfigDataError(
                f"Invalid {CONTEXTS_FIELD}, should be a list"
            )
        self.current_context = config_data[CURR_CONTEXT_FIELD]
        self.contexts: Dict[str, Context] = {}
        self.context_paths: Dict[str, str] = {}

    def add_context(self, context_data: Dict):
        context = Context(context_data)
        self.contexts[context.name] = context
        self.context_paths[context.name] = self.config_path

    def merge(self, new_config: "Config"):
        """Merges new_config into existing config overwriting data with
        matching keys

        Args:
            new_config (Config): a more-local config to merge into the
                current config.
        """
        for context_name, context in new_config.contexts.items():
            self.contexts[context_name] = context
            self.context_paths[context_name] = new_config.context_paths[
                context_name
            ]
        self.current_context = new_config.current_context
        self.config_dir = new_config.config_dir
        self.config_path = new_config.config_path

    def as_dict(self) -> Dict:
        rv = {
            lib.API_FIELD: lib.API_VERSION,
            lib.KIND_FIELD: CONFIG_KIND,
            CONTEXTS_FIELD: [
                context.as_dict() for context in self.contexts.values()
            ],
            CURR_CONTEXT_FIELD: self.current_context,
        }
        return rv

    def is_valid(self) -> bool:
        if len(self.validate()) > 0:
            return False
        else:
            return True

    def validate(self) -> List[str]:
        pass


def load_config(silent=False):
    """Loads spyctl configurations from disk starting with the current
    directory and ending with the global config. If no configuration
    file exists, the program exits.
    """
    if not GLOBAL_CONFIG_PATH.exists():
        GLOBAL_CONFIG_DIR.mkdir(parents=True, exist_ok=True)
        GLOBAL_CONFIG_PATH.touch()
        with GLOBAL_CONFIG_PATH.open("w") as f:
            yaml.dump(CONFIG_TEMPLATE, f)
    if not GLOBAL_SECRETS_PATH.exists():
        GLOBAL_SECRETS_DIR.mkdir(parents=True, exist_ok=True)
        # GLOBAL_SECRETS_DIR.chmod(700)
        GLOBAL_SECRETS_PATH.touch()
        # GLOBAL_SECRETS_PATH.chmod(600)
    global LOADED_CONFIG, CURRENT_CONTEXT, LOADED_CONFIGS
    if LOADED_CONFIG is None or TESTING:
        LOADED_CONFIG = None
        LOADED_CONFIGS = {}
        s.load_secrets()
        configs_data = lib.walk_up_tree(GLOBAL_CONFIG_PATH, LOCAL_CONFIG_PATH)
        configs: List[Config] = []
        seen = set()
        for config_path, config_data in configs_data:
            if not schemas.valid_object(config_data):
                cli.err_exit(f"Config at {str(config_path)} is invalid.")
            try:
                cfg = Config(config_data, config_path, config_path.parent)
                cfg_cpy = Config(
                    deepcopy(config_data), config_path, config_path.parent
                )
                for context_data in config_data[lib.CONTEXTS_FIELD]:
                    if not schemas.valid_context(context_data):
                        ctx_name = context_data.get(lib.CONTEXT_NAME_FIELD)
                        prefix = (
                            f"Context {ctx_name!r}" if ctx_name else "Context"
                        )
                        cli.try_log(f"{prefix} in {config_path} is invalid")
                        continue
                    cfg.add_context(context_data)
                    cfg_cpy.add_context(context_data)
                configs.append(cfg)
                LOADED_CONFIGS[str(config_path)] = deepcopy(cfg_cpy)
            except InvalidConfigDataError as e:
                if str(config_path) not in seen and not silent:
                    cli.try_log(
                        f"Config at {str(config_path)} is invalid."
                        f" {' '.join(e.args)}"
                    )
            except InvalidContextDataError as e:
                if str(config_path) not in seen and not silent:
                    cli.try_log(
                        f"Config at {str(config_path)} has an invalid context."
                        f" {' '.join(e.args)}"
                    )
            seen.add(str(config_path))
        if len(configs) == 0 and not silent:
            cli.try_log(
                "No valid configurations."
                " Clear or fix any invalid configuration files."
            )
        else:
            base_config = configs.pop()
            # Reversed so that most local config takes priority
            for config in reversed(configs):
                base_config.merge(config)
            LOADED_CONFIG = base_config
            set_current_context(
                base_config.contexts.get(base_config.current_context)
            )


def get_loaded_config() -> Optional[Config]:
    return LOADED_CONFIG


def set_current_context(ctx: Optional[Context]):
    global CURRENT_CONTEXT
    CURRENT_CONTEXT = ctx


def clear_current_context():
    global CURRENT_CONTEXT
    CURRENT_CONTEXT = None


def get_current_context() -> Context:
    if CURRENT_CONTEXT:
        return CURRENT_CONTEXT
    config = get_loaded_config()
    secrets = s.get_secrets()
    if len(secrets) == 0:
        cli.try_log(
            "No secrets found. Create a secret for accessing the spyderbat api"
            " via the command 'spyctl create secret apicfg'.",
            is_warning=True,
        )
    if len(config.contexts) == 0:
        cli.err_exit(
            "No valid contexts. Try using 'spyctl config set-context' to"
            " create one."
        )
    if (
        config.current_context == CURR_CONTEXT_NONE
        or not config.current_context
    ):
        cli.err_exit(
            "Current context is not set. Try using"
            " 'spyctl config use-context'"
        )
    if config.current_context not in config.contexts:
        cli.err_exit(
            "Unable to locate current context"
            f" '{config.current_context}'."
            " Try using 'spyctl config use-context'."
        )
    if CURRENT_CONTEXT is None:
        cli.err_exit("No current context set")
    cli.err_exit("Configuration not complete.")


def init():
    perform_init = False
    reset = False
    local_workspace_path = Path.joinpath(Path.cwd(), LOCAL_CONFIG_PATH)
    if (
        local_workspace_path.exists()
        and local_workspace_path == GLOBAL_CONFIG_PATH
    ):
        if cli.query_yes_no(
            "Are you sure you want to reset the global spyctl configuration"
            f" file '{str(local_workspace_path)}'",
            "no",
        ):
            perform_init = True
            reset = True
    elif local_workspace_path.exists():
        if cli.query_yes_no(
            "A spyctl configuration file already exists at"
            f" '{str(local_workspace_path)}'. Are you sure"
            " you want to reset it?",
            "no",
        ):
            reset = True
            perform_init = True
    else:
        perform_init = True
    if perform_init:
        global_config = LOADED_CONFIGS.get(str(GLOBAL_CONFIG_PATH))
        global_current_ctx = None
        if (
            global_config
            and global_config.current_context
            and global_config.current_context != CURR_CONTEXT_NONE
        ):
            global_current_ctx = global_config.current_context
        local_workspace_path.parent.mkdir(parents=True, exist_ok=True)
        with local_workspace_path.open("w") as f:
            config_template = deepcopy(CONFIG_TEMPLATE)
            if global_current_ctx:
                config_template[CURR_CONTEXT_FIELD] = global_current_ctx
            yaml.dump(config_template, f)
        cli.try_log(
            f"{'Reset' if reset else 'Created'} configuration file at"
            f" {str(local_workspace_path)}."
        )


def set_context(
    name, secret, force_global: bool, use_context: bool, **context
):
    global LOADED_CONFIGS
    new_context = {
        CONTEXT_NAME_FIELD: name,
        SECRET_FIELD: secret,
        CONTEXT_FIELD: context,
    }
    try:
        new_context = Context(new_context)
        # Initiate a validation of the org
        new_context.get_api_data()
    except InvalidContextDataError as e:
        cli.err_exit(f"Unable to set context. {' '.join(e.args)}")
    config = get_loaded_config()
    if config is None:
        return
    updated = name in config.contexts
    if force_global:
        # global flag was set so update/set context at the global level
        context_path = GLOBAL_CONFIG_PATH
    else:
        if updated:
            # update the context at the file of the existing context
            context_path = config.context_paths[name]
        else:
            # set the context at the most-local configuration file
            context_path = config.config_path
    target_config = LOADED_CONFIGS[str(context_path)]
    tgt_updated_curr_ctx = False
    if len(target_config.contexts) == 0:
        target_config.current_context = name
        tgt_updated_curr_ctx = True
    # We want to ensure the current-context of the local configuration
    # file has been set. local config can be the same as target or not,
    # but it is the configuration file closest to the user's cwd
    local_config = LOADED_CONFIGS[str(config.config_path)]
    local_updated_curr_ctx = False
    if (
        len(config.contexts) == 0
        and len(local_config.contexts) == 0
        and local_config.config_path != context_path
    ) or use_context:
        local_updated_curr_ctx = True
        local_config.current_context = name
    target_config.contexts[name] = new_context
    target_config.context_paths[name] = config.config_path
    try:
        with context_path.open("w") as f:
            yaml.dump(target_config.as_dict(), f, sort_keys=False)
            cli.try_log(
                f"{'Updated' if updated else 'Set new'} context '{name}' in"
                f" configuration file '{context_path}'."
            )
            if tgt_updated_curr_ctx:
                cli.try_log(
                    f"Updated current context to '{name}' in"
                    f" configuration file '{context_path}'."
                )
        if local_updated_curr_ctx:
            with local_config.config_path.open("w") as f:
                yaml.dump(local_config.as_dict(), f, sort_keys=False)
                if (
                    local_config.config_path != context_path
                    or not tgt_updated_curr_ctx
                ):
                    cli.try_log(
                        f"Updated current context to '{name}' in"
                        f" configuration file '{local_config.config_path}'."
                    )
        if name != local_config.current_context:
            cli.try_log(
                "NOTICE: The context you just set is not your current context."
                f" Use 'spyctl config use-context {name}' to set it as your"
                " current context."
            )
    except Exception as e:
        cli.err_exit(f"Unable to set context. {' '.join(e.args)}")


def delete_context(name, force_global):
    global LOADED_CONFIGS
    config = get_loaded_config()
    if config is None:
        cli.err_exit("Unable to load configuration")
    if force_global:
        # global flag was set so delete context at the global level
        context_path = GLOBAL_CONFIG_PATH
        target_config = LOADED_CONFIGS.get(str(GLOBAL_CONFIG_PATH))
    else:
        # delete the context in the file where it resides
        context_path = config.context_paths[name]
        target_config = LOADED_CONFIGS.get(str(context_path))
    if not target_config:
        cli.err_exit("Unable to load target configuration")
    if name not in target_config.contexts:
        cli.err_exit(
            f"Unable to delete context '{name}' from {str(context_path)}."
            " Context does not exist in target config."
        )
    del target_config.contexts[name]
    updated_curr_context = False
    if target_config.current_context == name:
        if len(target_config.contexts) > 0:
            fallback_context = next(iter(target_config.contexts.values()))
            target_config.current_context = fallback_context.name
            updated_curr_context = True
        elif len(config.contexts) > 0:
            fallback_context = next(iter(config.contexts.values()))
            target_config.current_context = fallback_context.name
            updated_curr_context = True
        else:
            target_config.current_context = CURR_CONTEXT_NONE
    try:
        with context_path.open("w") as f:
            yaml.dump(target_config.as_dict(), f, sort_keys=False)
            cli.try_log(
                f"Deleted context '{name}' in"
                f" configuration file '{context_path}'."
            )
            if updated_curr_context:
                cli.try_log(
                    f"Current context at {target_config.config_path} deleted."
                    " Updating to fallback context"
                    f" '{target_config.current_context}'."
                )
            if len(config.contexts) == 0:
                cli.try_log(
                    "Configuration has no contexts. Use 'spyctl config"
                    " set-context' to add one."
                )
    except Exception as e:
        cli.err_exit(f"Unable to delete context. {' '.join(e.args)}")


def current_config():
    global LOADED_CONFIGS
    config = get_loaded_config()
    if not config:
        cli.err_exit("Unable to load config.")
    cli.try_print(config.config_path)


def current_context(force_global):
    global LOADED_CONFIGS
    if force_global:
        config = LOADED_CONFIGS.get(str(GLOBAL_CONFIG_PATH))
    else:
        config = get_loaded_config()
    if not config:
        cli.err_exit("Unable to load config.")
    if (
        config.current_context == CURR_CONTEXT_NONE
        or not config.current_context
    ):
        if len(config.contexts) > 0:
            cli.try_log(
                "No current context. Use 'spyctl config use-context' to set"
                " one."
            )
        elif len(config.contexts) > 0:
            cli.try_log(
                "No current context. Use 'spyctl config set-context' to create"
                " and set one."
            )
    else:
        cli.try_print(config.current_context)


def use_context(name, force_global):
    global LOADED_CONFIGS
    if force_global:
        # global flag was set so update/set context at the global level
        config = LOADED_CONFIGS.get(str(GLOBAL_CONFIG_PATH))
        context_path = GLOBAL_CONFIG_PATH
    else:
        # update the current context in the local configuration file
        config = get_loaded_config()
        context_path = config.config_path
    if not config:
        cli.err_exit("Unable to load config.")
    found = name in config.contexts
    if not found:
        cli.try_log(
            f"Unable to set current context '{name}' for {str(context_path)}."
            " Context does not exist in target config."
        )
        return
    target_config = LOADED_CONFIGS[str(context_path)]
    target_config.current_context = name
    try:
        with context_path.open("w") as f:
            yaml.dump(target_config.as_dict(), f, sort_keys=False)
            cli.try_log(
                f"Set current context to '{name}' in"
                f" configuration file '{context_path}'."
            )
    except Exception as e:
        cli.err_exit(f"Unable to delete context. {' '.join(e.args)}")


def get_contexts(name, force_global, force_workspace, output):
    if force_global:
        # global flag was set so update/set context at the global level
        config = LOADED_CONFIGS.get(str(GLOBAL_CONFIG_PATH))
    elif force_workspace:
        # Get all loaded config paths
        config_paths = list(LOADED_CONFIGS)
        if len(config_paths) > 0:
            # The most-local config will be the first one in the list
            # of config paths
            config = LOADED_CONFIGS[config_paths[0]]
        else:
            config = None
    else:
        config = get_loaded_config()
    # contexts = [ctx.as_dict() for ctx in config.contexts.values()]
    contexts = []
    if not config:
        cli.err_exit("Unable to load config.")
    for context in config.contexts.values():
        ctx_dict = context.as_dict()
        ctx_dict[lib.LOCATION_FIELD] = config.context_paths.get(
            context.name, "Unknown"
        )
        contexts.append(ctx_dict)
    if name:
        contexts = filter(lambda x: x[lib.NAME_FIELD] == name, contexts)
    cli.show(
        (contexts, config.current_context),
        output,
        {
            lib.OUTPUT_WIDE: context_wide_output,
            lib.OUTPUT_DEFAULT: context_summary_output,
        },
    )


def context_summary_output(contexts_tup: Tuple[List[Dict], str]) -> str:
    contexts, current = contexts_tup
    headers = ["CURRENT", "NAME", "ORGANIZATION", "FILTERS"]
    data = []
    for context in contexts:
        name = context[lib.NAME_FIELD]
        org = context[CONTEXT_FIELD][ORG_FIELD]
        len_ctx_filters_minus_org = len(context[CONTEXT_FIELD]) - 1
        data.append(
            [
                "*" if name == current else "",
                name,
                org,
                len_ctx_filters_minus_org,
            ]
        )
    data.sort(key=lambda x: x[1])
    return tabulate(data, headers, tablefmt="plain")


def context_wide_output(contexts_tup: Tuple[List[Dict], str]) -> str:
    contexts, current = contexts_tup
    headers = ["CURRENT", "NAME", "ORGANIZATION", "LOCATION", "FILTERS"]
    data = []
    for context in contexts:
        name = context[lib.NAME_FIELD]
        org = context[CONTEXT_FIELD][ORG_FIELD]
        filters: Dict = context[CONTEXT_FIELD].copy()
        filters.pop(ORG_FIELD)
        data.append(
            [
                "*" if name == current else "",
                name,
                org,
                context[lib.LOCATION_FIELD],
                "\n".join(
                    [f"{key}: {value}" for key, value in filters.items()]
                ),
            ]
        )
    data.sort(key=lambda x: x[1])
    return tabulate(data, headers, tablefmt="plain")


def view_config(force_global, force_workspace, output):
    global LOADED_CONFIGS
    config = get_loaded_config()
    if config is None:
        return
    if force_global:
        # global flag was set so update/set context at the global level
        cfg_to_view = LOADED_CONFIGS[str(GLOBAL_CONFIG_PATH)].as_dict()
    elif force_workspace:
        cfg_to_view = LOADED_CONFIGS[str(config.config_path)].as_dict()
    else:
        cfg_to_view = config.as_dict()
    if output == lib.OUTPUT_DEFAULT:
        output = lib.OUTPUT_YAML
    cli.show(cfg_to_view, output)


# def validate_config(config_data: dict):


def validate_org(org, api_url, api_key) -> Optional[str]:
    orgs = api.get_orgs(api_url, api_key)
    if orgs is not None:
        for uid, name in zip(*orgs):
            if org == name or org == uid:
                if name == "Defend The Flag":
                    cli.err_exit("invalid organization")
                return uid
    return None


def set_testing(workspace_path=None):
    global TESTING
    TESTING = True


class ContextsParam(click.ParamType):
    name = "contexts_param"

    def shell_complete(self, ctx, param, incomplete):
        load_config(silent=True)
        config = get_loaded_config()
        if config:
            context_names = [
                context.name for context in config.contexts.values()
            ]
            context_names.sort()
            return [
                CompletionItem(context_name)
                for context_name in context_names
                if context_name.startswith(incomplete)
            ]
        else:
            return []


def set_api_call():
    s.set_api_call()
    lib.set_api_call()
    cli.set_yes_option()


def use_temp_secret_and_context(org_uid, api_key, api_url):
    context = create_temp_secret_and_context(org_uid, api_key, api_url)
    set_current_context(context)


def create_temp_secret_and_context(org_uid, api_key, api_url):
    secret_name = "__temp_secret__"
    context_name = "__temp_context"
    secret_data = {
        lib.API_FIELD: lib.API_VERSION,
        lib.KIND_FIELD: lib.SECRET_KIND,
        lib.METADATA_FIELD: {
            lib.METADATA_NAME_FIELD: secret_name,
        },
        lib.STRING_DATA_FIELD: {
            lib.API_KEY_FIELD: api_key,
            lib.API_URL_FIELD: api_url,
        },
    }
    context_data = {
        lib.CONTEXT_NAME_FIELD: context_name,
        lib.SECRET_FIELD: secret_name,
        lib.CONTEXT_FIELD: {lib.ORG_FIELD: org_uid},
    }
    secret = s.Secret(secret_data)
    s.SECRETS[secret_name] = secret
    context = Context(context_data)
    return context



import fnmatch
import time
from copy import deepcopy
from typing import Dict, Iterable, List, Optional, Union

import spyctl.config.configs as cfgs
import spyctl.spyctl_lib as lib

DEFAULT_FILTER_TIME = (lib.time_inp("2h"), time.time())

CONT_SEL_TGT = [lib.SPEC_FIELD, lib.CONT_SELECTOR_FIELD]
SVC_SEL_TGT = [lib.SPEC_FIELD, lib.SVC_SELECTOR_FIELD]
NS_SEL_TGT = [
    lib.SPEC_FIELD,
    lib.NAMESPACE_SELECTOR_FIELD,
    lib.MATCH_LABELS_FIELD,
]
POD_SEL_TGT = [lib.SPEC_FIELD, lib.POD_SELECTOR_FIELD, lib.MATCH_LABELS_FIELD]
CLUSTERS_TGT_FIELDS = ["uid", "name"]
MACHINES_TGT_FIELDS = ["uid", "name"]
CGROUP_TGT_FIELDS = [[*SVC_SEL_TGT, lib.CGROUP_FIELD]]
CONT_NAME_TGT_FIELDS = [*CONT_SEL_TGT, lib.CONT_NAME_FIELD]
CONT_ID_TGT_FIELDS = [[*CONT_SEL_TGT, lib.CONT_ID_FIELD]]
IMAGE_TGT_FIELDS = [[*CONT_SEL_TGT, lib.IMAGE_FIELD]]
IMAGEID_TGT_FIELDS = [[*CONT_SEL_TGT, lib.IMAGEID_FIELD]]
NAMESPACE_LABEL_TGT_FIELDS = [[*NS_SEL_TGT]]
POD_LABEL_TGT_FIELDS = [[*POD_SEL_TGT]]


def filter_clusters(
    clusters_data: List[Dict],
    **filters,
):
    filter_set = {
        cfgs.CLUSTER_FIELD: lambda data, filt: filter_obj(
            data, CLUSTERS_TGT_FIELDS, filt
        ),
    }
    clusters_data = use_filters(clusters_data, filter_set, filters)
    return clusters_data


def filter_deployments(
    deployments_data: List[Dict],
    clusters_data=None,
    machines_data=None,
    pods_data=None,
    cgroups_data=None,
    containers_data=None,
    **filters,
):
    filter_set = {
        cfgs.NAMESPACE_FIELD: lambda data, filt: filter_obj(
            data, [[lib.METADATA_FIELD, lib.NAMESPACE_FIELD]], filt
        ),
    }
    deployments_data = use_filters(deployments_data, filter_set, filters)
    return deployments_data


def filter_namespaces(
    namespaces_data: List[Dict],
    clusters_data=None,
    machines_data=None,
    pods_data=None,
    cgroups_data=None,
    containers_data=None,
    **filters,
):
    filter_set = {
        cfgs.CLUSTER_FIELD: lambda data, filt: filter_obj(
            data, ["cluster_uid", "cluster_name"], filt
        ),
    }
    namespaces_data = use_filters(namespaces_data, filter_set, filters)
    return namespaces_data


def filter_spydertraces(
    spydertraces_data: List[Dict],
    clusters_data=None,
    machines_data=None,
    pods_data=None,
    cgroups_data=None,
    containers_data=None,
    **filters,
):
    return spydertraces_data


def filter_sources(
    sources_data: List[Dict],
    use_context_filters=True,
    **filters,
):
    filter_set = {
        cfgs.MACHINES_FIELD: lambda data, filt: filter_obj(
            data, MACHINES_TGT_FIELDS, filt
        ),
    }
    sources_data = use_filters(
        sources_data, filter_set, filters, use_context_filters
    )
    return sources_data


def filter_nodes(
    nodes_data: List[Dict],
    clusters_data=None,
    namespaces_data=None,
    pods_data=None,
    cgroups_data=None,
    containers_data=None,
    **filters,
):
    filter_set = {
        cfgs.MACHINES_FIELD: lambda data, filt: filter_obj(
            data, ["muid"], filt
        ),
    }
    nodes_data = use_filters(nodes_data, filter_set, filters)
    return nodes_data


def filter_redflags(
    flag_grp_data: List[Dict],
    namespaces_data=None,
    clusters_data=None,
    machines_data=None,
    pods_data=None,
    cgroups_data=None,
    containers_data=None,
    **filters,
):
    def severity_filter(data, filt):
        index = -1
        try:
            index = lib.ALLOWED_SEVERITIES.index(filt)
        except ValueError:
            return data
        rv = []
        for flag in data:
            try:
                if lib.ALLOWED_SEVERITIES.index(flag["severity"]) <= index:
                    rv.append(flag)
            except ValueError:
                rv.append(flag)
        return rv

    def exceptions_filter(data, filt):
        if filt:
            return data
        return [flag for flag in data if not flag["false_positive"]]

    filter_set = {
        cfgs.MACHINES_FIELD: lambda data, filt: filter_obj(
            data, ["muid"], filt
        ),
        lib.FLAG_SEVERITY: severity_filter,
        "exceptions": exceptions_filter,
    }
    flag_grp_data = use_filters(flag_grp_data, filter_set, filters)
    return flag_grp_data


def filter_opsflags(
    flag_grp_data: List[Dict],
    namespaces_data=None,
    clusters_data=None,
    machines_data=None,
    pods_data=None,
    cgroups_data=None,
    containers_data=None,
    **filters,
):
    def severity_filter(data, filt):
        index = -1
        try:
            index = lib.ALLOWED_SEVERITIES.index(filt)
        except ValueError:
            return data
        rv = []
        for flag in data:
            try:
                if lib.ALLOWED_SEVERITIES.index(flag["severity"]) <= index:
                    rv.append(flag)
            except ValueError:
                rv.append(flag)
        return rv

    filter_set = {
        cfgs.MACHINES_FIELD: lambda data, filt: filter_obj(
            data, ["muid"], filt
        ),
        lib.FLAG_SEVERITY: severity_filter,
    }
    flag_grp_data = use_filters(flag_grp_data, filter_set, filters)
    return flag_grp_data


def filter_fingerprints(
    fingerprint_data: List[Dict],
    namespaces_data=None,
    clusters_data=None,
    machines_data=None,
    pods_data=None,
    cgroups_data=None,
    containers_data=None,
    use_context_filters=True,
    suppress_warning=False,
    not_matching=False,
    **filters,
):
    def cont_id_filter(data, filt):
        filt += "*" if filt[-1] != "*" else filt
        return filter_obj(data, CONT_ID_TGT_FIELDS, filt)

    def image_id_filter(data, filt):
        filt += "*" if filt[-1] != "*" else filt
        return filter_obj(data, IMAGEID_TGT_FIELDS, filt)

    def filter_namespace_labels(data, filt: Dict):
        for key, value in filt.items():
            tgt_fields = deepcopy(NAMESPACE_LABEL_TGT_FIELDS)
            tgt_fields[0].append(key)
            data = filter_obj(
                data,
                tgt_fields,
                value,
            )
        return data

    def filter_pod_labels(data, filt: Dict):
        for key, value in filt.items():
            tgt_fields = deepcopy(POD_LABEL_TGT_FIELDS)
            tgt_fields[0].append(key)
            data = filter_obj(
                data,
                tgt_fields,
                value,
            )
        return data

    filter_set = {
        cfgs.CGROUP_FIELD: lambda data, filt: filter_obj(
            data, CGROUP_TGT_FIELDS, filt
        ),
        lib.CONT_NAME_FIELD: lambda data, filt: filter_obj(
            data, CONT_NAME_TGT_FIELDS, filt
        ),
        lib.CONT_ID_FIELD: cont_id_filter,
        lib.IMAGE_FIELD: lambda data, filt: filter_obj(
            data, IMAGE_TGT_FIELDS, filt
        ),
        lib.IMAGEID_FIELD: image_id_filter,
        lib.NAMESPACE_FIELD: lambda data, filt: filter_obj(
            data,
            [[lib.METADATA_FIELD, lib.METADATA_NAMESPACE_FIELD]],
            filt,
        ),
        lib.NAMESPACE_LABELS_FIELD: filter_namespace_labels,
        lib.POD_LABELS_FIELD: filter_pod_labels,
    }
    if not_matching:
        non_matches = []
        for fingerprint in fingerprint_data:
            if not use_filters(
                [fingerprint],
                filter_set,
                filters,
                use_context_filters,
                suppress_warning=True,
            ):
                non_matches.append(fingerprint)
        fingerprint_data = non_matches
    else:
        fingerprint_data = use_filters(
            fingerprint_data,
            filter_set,
            filters,
            use_context_filters,
            suppress_warning=suppress_warning,
        )
    return fingerprint_data


def filter_policies(
    policy_data: List[Dict],
    namespaces_data=None,
    clusters_data=None,
    machines_data=None,
    pods_data=None,
    cgroups_data=None,
    containers_data=None,
    **filters,
):
    return policy_data


def filter_pods(
    pods_data: List[Dict],
    namespaces_data=None,
    clusters_data=None,
    machines_data=None,
    cgroups_data=None,
    containers_data=None,
    **filters,
):
    filter_set = {
        cfgs.NAMESPACE_FIELD: lambda data, filt: filter_obj(
            data, [[lib.METADATA_FIELD, lib.METADATA_NAMESPACE_FIELD]], filt
        ),
    }
    pods_data = use_filters(pods_data, filter_set, filters)
    return pods_data


def filter_processes(
    processes: List[Dict],
    namespaces_data=None,
    clusters_data=None,
    machines_data=None,
    pods_data=None,
    cgroups_data=None,
    containers_data=None,
    **filters,
):
    return processes


def filter_containers(
    containers: List[Dict],
    namespaces_data=None,
    clusters_data=None,
    machines_data=None,
    pods_data=None,
    cgroups_data=None,
    containers_data=None,
    **filters,
):
    return containers


def filter_connections(
    connections: List[Dict],
    namespaces_data=None,
    clusters_data=None,
    machines_data=None,
    pods_data=None,
    cgroups_data=None,
    containers_data=None,
    **filters,
):
    return connections


def use_filters(
    data,
    filter_functions: Dict,
    filters: Dict,
    use_context_filters=True,
    suppress_warning=False,
):
    ctx_filters = cfgs.get_current_context().get_filters()
    data_empty_at_start = len(data) == 0
    for filt, func in filter_functions.items():
        if filt in filters:
            data = func(data, filters[filt])
        elif use_context_filters and filt in ctx_filters:
            data = func(data, ctx_filters[filt])
        if len(data) == 0 and not data_empty_at_start and not suppress_warning:
            lib.try_log(f"No results after filtering on '{filt}'")
            return data
    return data


def filter_obj(
    obj: List[Dict],
    target_fields: List[Union[str, List[str]]],
    filters: Union[List[str], str],
) -> List[Dict]:
    rv = []
    if "-all" in filters:
        return obj
    if isinstance(filters, list):
        for rec in obj:
            if match_filters(rec, target_fields, filters):
                rv.append(rec)
    else:
        for rec in obj:
            if match_filters(rec, target_fields, [filters]):
                rv.append(rec)
    return rv


def match_filters(
    record: Dict, target_fields: List[str], filters: List[str]
) -> bool:
    for fil in filters:
        for field in target_fields:
            value = get_field_value(field, record)
            if value is None:
                continue
            if "*" in fil:
                if isinstance(value, str) and fnmatch.fnmatch(value, fil):
                    return True
                try:
                    if not isinstance(value, str):
                        for val in get_field_value(field, record):
                            if fnmatch.fnmatch(val, fil):
                                return True
                except Exception:
                    pass
            else:
                if value == fil:
                    return True
                try:
                    if fil in get_field_value(
                        field, record
                    ) and not isinstance(value, str):
                        return True
                except Exception:
                    pass
    return False


def filter_agents(
    agents: List[Dict],
    namespaces_data=None,
    clusters_data=None,
    machines_data=None,
    pods_data=None,
    cgroups_data=None,
    containers_data=None,
    **filters,
):
    return agents


def get_field_value(
    field: Union[str, List[str]], obj: Dict
) -> Optional[Union[str, Iterable]]:
    value = obj
    if isinstance(field, str):
        field = field.split(".")
    keys = field
    for key in keys:
        value = value.get(key)
        if value is None:
            return None
    return value



from typing import Dict, List, Tuple

from tabulate import tabulate

import spyctl.api as api
import spyctl.config.configs as cfg
import spyctl.spyctl_lib as lib

SUMMARY_HEADERS = ["NAME", "UID", "OS", "CLOUD_TYPE", "AGE", "CLUSTER"]


def machines_summary_output(
    ctx: cfg.Context,
    muids: List[str],
    time: Tuple[float, float],
    pipeline=None,
    limit_mem=False,
) -> str:
    data = []
    for machine in api.get_machines(
        *ctx.get_api_data(), muids, time, pipeline, limit_mem
    ):
        data.append(__machine_summary_data(machine))
    data.sort(key=lambda x: [x[0], lib.to_timestamp(x[3])])
    return tabulate(data, SUMMARY_HEADERS, tablefmt="plain")


def __machine_summary_data(machine: Dict):
    cloud_type = machine.get("cloud_type")
    if not cloud_type:
        cloud_type = lib.NOT_AVAILABLE
    cluster = machine.get("cluster_name")
    if not cluster:
        cluster = lib.NOT_AVAILABLE
    rv = [
        machine[lib.HOSTNAME_FIELD],
        machine[lib.ID_FIELD],
        machine["os_version"],
        cloud_type,
        lib.calc_age(machine["boot_time"]),
        cluster,
    ]
    return rv



from typing import Dict, List, Tuple

from tabulate import tabulate

import spyctl.api as api
import spyctl.config.configs as cfg
import spyctl.spyctl_lib as lib

NOT_AVAILABLE = lib.NOT_AVAILABLE
SUMMARY_HEADERS = [
    "FLAG",
    "SEVERITY",
    "COUNT",
    "LATEST_TIMESTAMP",
    "REF_OBJ",
]


class FlagsGroup:
    def __init__(self) -> None:
        self.ref_flag = None
        self.latest_timestamp = NOT_AVAILABLE
        self.count = 0

    def add_flag(self, flag: Dict):
        if self.ref_flag is None:
            self.ref_flag = flag
        self.__update_latest_timestamp(flag.get("time"))
        self.count += 1

    def __update_latest_timestamp(self, timestamp):
        if timestamp is None:
            return
        if (
            self.latest_timestamp is None
            or self.latest_timestamp == NOT_AVAILABLE
        ):
            self.latest_timestamp = timestamp
        elif timestamp > self.latest_timestamp:
            self.latest_timestamp = timestamp

    def summary_data(self) -> List[str]:
        timestamp = NOT_AVAILABLE
        if (
            self.latest_timestamp is not None
            and self.latest_timestamp != NOT_AVAILABLE
        ):
            timestamp = lib.epoch_to_zulu(self.latest_timestamp)
        ref_obj = self.ref_flag["class"][1]
        if ref_obj in lib.CLASS_LONG_NAMES:
            ref_obj = lib.CLASS_LONG_NAMES[ref_obj]
        rv = [
            self.ref_flag["short_name"],
            self.ref_flag["severity"],
            str(self.count),
            timestamp,
            ref_obj,
        ]
        return rv


def flags_output_summary(
    ctx: cfg.Context,
    flag_type: str,
    muids: List[str],
    time: Tuple[float, float],
    pipeline=None,
    limit_mem=False,
) -> str:
    groups: Dict[str, FlagsGroup] = {}
    if flag_type == lib.EVENT_OPSFLAG_PREFIX:
        api_func = api.get_opsflags
    else:
        api_func = api.get_redflags
    for flag in api_func(
        *ctx.get_api_data(), muids, time, pipeline, limit_mem
    ):
        flag_class = "/".join(flag["class"])
        if flag_class not in groups:
            groups[flag_class] = FlagsGroup()
        groups[flag_class].add_flag(flag)
    data = []
    for group in groups.values():
        data.append(group.summary_data())
    output = tabulate(
        sorted(
            data,
            key=lambda x: [
                _severity_index(x[1]),
                x[0],
                x[4],
                lib.to_timestamp(x[3]),
            ],
        ),
        headers=SUMMARY_HEADERS,
        tablefmt="plain",
    )
    return output


def _severity_index(severity):
    try:
        return lib.ALLOWED_SEVERITIES.index(severity)
    except ValueError:
        return -1



import json
import tempfile
from io import TextIOWrapper
from typing import IO, Callable, Dict

import click
import yaml

import spyctl.api as api
import spyctl.cli as cli
import spyctl.config.configs as cfg
import spyctl.filter_resource as filt
import spyctl.resources.policies as p
import spyctl.resources.suppression_policies as sp
import spyctl.resources.notification_targets as nt
import spyctl.resources.notifications_configs as nc
import spyctl.schemas_v2 as schemas
import spyctl.spyctl_lib as lib

EDIT_PROMPT = (
    "# Please edit the object below. Lines beginning with a '#' will be ignored,\n"
    "# and an empty file will abort the edit. If an error occurs while saving this file will be\n"
    "# reopened with the relevant failures.\n"
    "#\n"
)


def handle_edit(resource=None, name_or_id=None, file: IO = None):
    """
    Handles the 'edit' command for a specified resource and name or id.

    Args:
        resource (str): The resource to edit.
        name_or_id (str): The name or id of the resource to edit.
        file (IO): The file to edit.

    Returns:
        None
    """
    if file:
        handle_edit_file(file)
    else:
        if not resource or not name_or_id:
            cli.err_exit("Must specify resource and name or id.")
        if resource == lib.NOTIFICATION_CONFIGS_RESOURCE:
            handle_edit_notif_config(name_or_id)
        elif resource == lib.NOTIFICATION_TARGETS_RESOURCE:
            handle_edit_notif_tgt(name_or_id)
        elif resource == lib.POLICIES_RESOURCE:
            handle_edit_policy(name_or_id)
        elif resource == lib.SUPPRESSION_POLICY_RESOURCE:
            handle_edit_suppression_policy(name_or_id)
        else:
            cli.err_exit(
                f"The 'edit' command is not supported for '{resource}'"
            )


KIND_TO_RESOURCE_TYPE: Dict[str, str] = {
    lib.BASELINE_KIND: lib.BASELINES_RESOURCE.name,
    lib.CONFIG_KIND: lib.CONFIG_ALIAS.name,
    lib.FPRINT_GROUP_KIND: lib.FINGERPRINT_GROUP_RESOURCE.name,
    lib.FPRINT_KIND: lib.FINGERPRINTS_RESOURCE.name,
    lib.POL_KIND: lib.POLICIES_RESOURCE.name,
    (lib.POL_KIND, lib.POL_TYPE_TRACE): lib.SUPPRESSION_POLICY_RESOURCE.name,
    lib.SECRET_KIND: lib.SECRETS_ALIAS.name,
    lib.UID_LIST_KIND: lib.UID_LIST_RESOURCE.name,
    lib.DEVIATION_KIND: lib.DEVIATIONS_RESOURCE.name,
    lib.NOTIFICATION_KIND: lib.NOTIFICATION_CONFIGS_RESOURCE.name,
    lib.TARGET_KIND: lib.NOTIFICATION_TARGETS_RESOURCE.name,
}


def handle_edit_file(file: IO):
    """
    Handle editing a file.

    Args:
        file (IO): The file to be edited.

    Raises:
        ValueError: If the file is not a valid resource file.
        ValueError: If editing a resource of the given kind is not supported.
    """
    resource = lib.load_resource_file(file)
    if not isinstance(resource, Dict):
        cli.err_exit("Invalid file for editing.")
    kind = resource.get(lib.KIND_FIELD)
    if kind not in KIND_TO_RESOURCE_TYPE:
        cli.err_exit(f"Editing resource of kind '{kind}' not supported.")
    if kind == lib.POL_KIND:
        type = resource[lib.METADATA_FIELD][lib.METADATA_TYPE_FIELD]
        if type == lib.POL_TYPE_TRACE:
            kind = (lib.POL_KIND, lib.POL_TYPE_TRACE)
    resource_type = KIND_TO_RESOURCE_TYPE[kind]
    edit_resource(
        yaml.dump(resource, sort_keys=False),
        file,
        resource_type,
        schemas.KIND_TO_SCHEMA[kind],
        apply_file_edits,
    )


def handle_edit_notif_config(name_or_id):
    """
    Edit a notification configuration based on the provided name or ID.

    Args:
        name_or_id (str): The name or ID of the notification configuration to
            edit.

    Raises:
        ValueError: If the provided name or ID is ambiguous or no matching
            notification configurations are found.
        ValueError: If editing legacy notification configurations is not
            supported.

    Returns:
        None
    """
    ctx = cfg.get_current_context()
    notif_pol = api.get_notification_policy(*ctx.get_api_data())
    routes = notif_pol.get(lib.ROUTES_FIELD)
    edit_index = None
    edit_id = None
    for i, route in enumerate(routes):
        id = route.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
        name = route.get(lib.DATA_FIELD, {}).get(lib.NAME_FIELD)
        if id == name_or_id or name == name_or_id:
            if edit_index is not None and name == name_or_id:
                cli.err_exit(f"{name_or_id} is ambiguous, use ID")
            edit_index = i
            edit_id = id
    if edit_index is None:
        cli.err_exit(f"No notification configs matching '{name_or_id}'.")
    config = (
        routes[edit_index]
        .get(lib.DATA_FIELD, {})
        .get(lib.NOTIF_SETTINGS_FIELD)
    )
    key = (
        lib.NOTIFICATION_KIND,
        config.get(lib.METADATA_FIELD).get(lib.METADATA_TYPE_FIELD),
    )
    if key not in schemas.KIND_TO_SCHEMA:
        key = lib.NOTIFICATION_KIND
    if not config:
        cli.err_exit("Editing legacy notification configs not supported.")
    resource_yaml = yaml.dump(config, sort_keys=False)
    edit_resource(
        resource_yaml,
        edit_id,
        lib.NOTIFICATION_CONFIGS_RESOURCE.name,
        schemas.KIND_TO_SCHEMA[key],
        apply_config_edits,
    )


def handle_edit_notif_tgt(name_or_id):
    """
    Handle the editing of a notification target based on the given name or ID.

    Args:
        name_or_id (str): The name or ID of the notification target to be
            edited.

    Returns:
        None
    """
    ctx = cfg.get_current_context()
    notif_pol = api.get_notification_policy(*ctx.get_api_data())
    targets: Dict = notif_pol[lib.TARGETS_FIELD]
    edit_tgt = None
    # check if name exists
    if name_or_id in targets:
        tgt_data = targets[name_or_id]
        edit_tgt = nt.Target(backend_target={name_or_id: tgt_data})
    if not edit_tgt:
        for tgt_name, tgt in targets.items():
            id = tgt.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
            if id is None:
                continue
            if id == name_or_id:
                edit_tgt = nt.Target(backend_target={tgt_name: tgt})
                break
    if not edit_tgt:
        cli.err_exit(f"No notification targets matching '{name_or_id}'.")
    resource_yaml = yaml.dump(edit_tgt.as_dict(), sort_keys=False)
    edit_resource(
        resource_yaml,
        edit_tgt.id,
        lib.NOTIFICATION_TARGETS_RESOURCE.name,
        schemas.NotificationTgtResourceModel,
        apply_tgt_edits,
    )


def handle_edit_policy(name_or_id):
    """
    Handle the editing of a policy based on the given name or ID.

    Args:
        name_or_id (str): The name or ID of the policy to be edited.

    Returns:
        None
    """
    ctx = cfg.get_current_context()
    policies = api.get_policies(*ctx.get_api_data())
    policies = filt.filter_obj(
        policies,
        [
            [lib.METADATA_FIELD, lib.NAME_FIELD],
            [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
        ],
        name_or_id,
    )
    if len(policies) > 1:
        cli.err_exit(f"Policy '{name_or_id}' is ambiguous, use ID.")
    if not policies:
        cli.err_exit(f"No Policies matching '{name_or_id}'.")
    policy = policies[0]
    resource_yaml = yaml.dump(policy, sort_keys=False)
    edit_resource(
        resource_yaml,
        policy[lib.METADATA_FIELD][lib.METADATA_UID_FIELD],
        lib.POLICIES_RESOURCE.name,
        schemas.KIND_TO_SCHEMA[lib.POL_KIND],
        apply_policy_edits,
    )


def handle_edit_suppression_policy(name_or_id):
    """
    Handle the editing of a suppression policy based on the given name or ID.

    Args:
        name_or_id (str): The name or ID of the suppression policy to be
            edited.

    Returns:
        None
    """
    ctx = cfg.get_current_context()
    policies = api.get_policies(
        *ctx.get_api_data(),
        params={lib.METADATA_TYPE_FIELD: lib.POL_TYPE_TRACE},
    )
    policies = filt.filter_obj(
        policies,
        [
            [lib.METADATA_FIELD, lib.NAME_FIELD],
            [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
        ],
        name_or_id,
    )
    if len(policies) > 1:
        cli.err_exit(
            f"Suppression Policy '{name_or_id}' is ambiguous, use ID."
        )
    if not policies:
        cli.err_exit(f"No Suppression Policies matching '{name_or_id}'.")
    policy = policies[0]
    resource_yaml = yaml.dump(policy, sort_keys=False)
    edit_resource(
        resource_yaml,
        policy[lib.METADATA_FIELD][lib.METADATA_UID_FIELD],
        lib.SUPPRESSION_POLICY_RESOURCE.name,
        schemas.KIND_TO_SCHEMA[(lib.POL_KIND, lib.POL_TYPE_TRACE)],
        apply_suppression_policy_edits,
    )


def edit_resource(
    resource_yaml: str,
    resource_id: str,
    resource_type,
    validator: Callable,
    apply_func: Callable,
):
    """
    Edit a resource using a YAML file.

    Args:
        resource_yaml (str): The YAML content of the resource.
        resource_id (str): The ID of the resource.
        resource_type: The type of the resource.
        validator (Callable): A function that validates the edited YAML
            content.
        apply_func (Callable): A function that applies the edited YAML content.

    Returns:
        None
    """
    temp_file = None
    while True:
        if not temp_file:
            edit_yaml = click.edit(
                __add_edit_prompt(resource_yaml), extension=".yaml"
            )
        else:
            try:
                with open(temp_file, "r") as f:
                    tmp_yaml = f.read()
                edit_yaml = click.edit(tmp_yaml, extension=".yaml")
            except Exception as e:
                cli.err_exit(str(e))
        if not edit_yaml or __strip_comments(edit_yaml) == resource_yaml:
            cli.try_log("Edit cancelled, no changes made.")
            exit(0)
        try:
            edit_dict = yaml.load(edit_yaml, lib.UniqueKeyLoader)
            error = None
            validator(**edit_dict)
        except Exception as e:
            error = str(e)
        if error and temp_file:
            edit_yaml = __add_error_comments(
                resource_type,
                edit_yaml,
                error,
                resource_id,
            )
            with open(temp_file, "w") as f:
                f.write(edit_yaml)
            cli.try_log(f"Edit failed, edits saved to {temp_file}")
            exit(1)
        if error:
            edit_yaml = __add_error_comments(
                resource_type,
                edit_yaml,
                error,
                resource_id,
            )
            temp_file = tempfile.NamedTemporaryFile(
                "w", delete=False, prefix="spyctl-edit-", suffix=".yaml"
            )
            temp_file.write(edit_yaml)
            cli.try_log(f"Edit failed, edits saved to {temp_file.name}")
            temp_file.close()
            temp_file = temp_file.name
            continue
        else:
            apply_func(edit_dict, resource_id)
            break


def apply_config_edits(edit_dict: Dict, config_id: str):
    """
    Apply edits to a notification configuration.

    Args:
        edit_dict (Dict): A dictionary containing the edits to be applied.
        config_id (str): The ID of the configuration to be edited.
    """
    ctx = cfg.get_current_context()
    notif_pol = api.get_notification_policy(*ctx.get_api_data())
    config = nc.NotificationConfig(edit_dict)
    config.id = config_id
    routes = notif_pol.get(lib.ROUTES_FIELD, [])
    edit_index = None
    for i, route in enumerate(routes):
        id = route.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
        if id == config_id:
            edit_index = i
            break
    if edit_index is None:
        cli.err_exit(f"Unable to locate config with id '{config_id}'")
    config.set_last_updated()
    routes[i] = config.route
    notif_pol[lib.ROUTES_FIELD] = routes
    api.put_notification_policy(*ctx.get_api_data(), notif_pol)
    cli.try_log(f"Successfully edited Notification Config '{config.id}'")


def apply_tgt_edits(edit_dict: Dict, target_id: str):
    """
    Apply edits to a notification target identified by target_id.

    Args:
        edit_dict (Dict): A dictionary containing the edited target resource.
        target_id (str): The ID of the target to be edited.

    Raises:
        ValueError: If the target with the specified ID is not found.

    """
    ctx = cfg.get_current_context()
    notif_pol = api.get_notification_policy(*ctx.get_api_data())
    target = nt.Target(target_resource=edit_dict)
    target.id = target_id
    targets: Dict = notif_pol.get(lib.TARGETS_FIELD, {})
    old_tgt = None
    for tgt_name, tgt_data in targets.items():
        tgt_id = tgt_data.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
        if not tgt_id:
            continue
        if tgt_id == target.id:
            old_tgt = {tgt_name: tgt_data}
            break
        if tgt_name == target.name:
            cli.err_exit("Target names must be unique!")
    if not old_tgt:
        cli.err_exit(f"Unable to locate config with id '{target_id}'")
    old_name = next(iter(old_tgt))
    targets.pop(old_name)
    target.set_last_update_time()
    new_tgt = target.as_target()
    targets.update(**new_tgt)
    notif_pol[lib.TARGETS_FIELD] = targets
    api.put_notification_policy(*ctx.get_api_data(), notif_pol)
    cli.try_log(f"Successfully edited Notification Target '{target.id}'")


def apply_policy_edits(edit_dict: Dict, policy_id: str):
    """
    Apply edits to a policy identified by policy_id.

    Args:
        edit_dict (Dict): A dictionary containing the edited policy resource.
        policy_id (str): The ID of the policy to be edited.
    """
    ctx = cfg.get_current_context()
    policy = p.Policy(edit_dict)
    _, api_data = p.get_data_for_api_call(policy)
    api.put_policy_update(*ctx.get_api_data(), policy_id, api_data)
    cli.try_log(f"Successfully edited Policy '{policy_id}'")


def apply_suppression_policy_edits(edit_dict: Dict, policy_id: str):
    ctx = cfg.get_current_context()
    policy = sp.TraceSuppressionPolicy(edit_dict)
    _, api_data = sp.get_data_for_api_call(policy)
    api.put_policy_update(*ctx.get_api_data(), policy_id, api_data)
    cli.try_log(f"Successfully edited Suppression Policy '{policy_id}'")


def apply_file_edits(resource, file: IO):
    """
    Apply edits to a file based on the given resource.

    Args:
        resource: The resource containing the edits.
        file (IO): The file to apply the edits to.

    """
    extension = ".json" if file.name.endswith(".json") else ".yaml"
    if extension == ".json":
        file.write(json.dumps(resource, sort_keys=False, indent=2))
    else:
        file.write(yaml.dump(resource, sort_keys=False))
    cli.try_log(f"Successfully edited resource file '{file.name}'")


def __strip_comments(yaml_string: str) -> str:
    lines = []
    for line in yaml_string.split("\n"):
        if line.strip().startswith("#"):
            continue
        lines.append(line)
    return "\n".join(lines)


def __add_edit_prompt(yaml_string: str):
    return EDIT_PROMPT + yaml_string


def __add_error_comments(
    resource: str, yaml_string: str, error: str, name: str = None
):
    if isinstance(name, TextIOWrapper):
        name = f' "{name.name}" '
    else:
        name = f' "{name}" ' if name else ""
    yaml_string = __strip_comments(yaml_string)
    error_prompt = f"# {resource}{name}was not valid:\n"
    error = error.split("\n")
    error = ["# " + line for line in error]
    error = "\n".join(error)
    error_prompt += error + "\n#\n"
    rv = EDIT_PROMPT + error_prompt + yaml_string
    return rv
